"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""

import logging
import sys

class ObjectFactory(object):
    def create_object(self, constr, named_constr):
        raise NotImplementedError()

class ReflectiveObjectFactory(ObjectFactory):
    def __init__(self, module_and_class):
        self.logger = logging.getLogger("springpython.factory.ReflectiveObjectFactory")
        self.module_and_class = module_and_class

    def create_object(self, constr, named_constr):
        self.logger.debug("Creating an instance of %s" % self.module_and_class)
        parts = self.module_and_class.split(".")
        module_name = ".".join(parts[:-1])
        class_name = parts[-1]
        if module_name == "":
            return __import__(class_name)(*constr, **named_constr)
        else:
            __import__(module_name)
            cls = getattr(sys.modules[module_name], class_name)
            return cls(*constr, **named_constr)


    def __str__(self):
        return "ReflectiveObjectFactory(%s)" % self.module_and_class

class PythonObjectFactory(ObjectFactory):
    def __init__(self, method, wrapper):
        self.logger = logging.getLogger("springpython.factory.PythonObjectFactory")
        self.method = method
        self.wrapper = wrapper

    def create_object(self, constr, named_constr):
        self.logger.debug("Creating an instance of %s" % self.method.func_name)
        
        # Setting wrapper's top_func can NOT be done earlier than this method call,
        # because it is tied to a wrapper decorator, which may not have yet been
        # generated.
        self.wrapper.func_globals["top_func"] = self.method.func_name
        
        # Because @object-based objects use direct code to specify arguments, and NOT
        # external configuration data, this factory doesn't care about the incoming arguments.
        
        return self.method()

    def __str__(self):
        return "PythonObjectFactory(%s)" % self.method



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""

# stdlib
import logging

# Circuits
from circuits import Component, Manager, Debugger

# ThreadPool
from threadpool import ThreadPool, WorkRequest, NoResultsPending

# Spring Python
from springpython.util import TRACE1
from springpython.context import InitializingObject
from springpython.jms import WebSphereMQJMSException, NoMessageAvailableException

class MessageHandler(object):
    def handle(self, message):
        raise NotImplementedError("Should be overridden by subclasses.")

class WebSphereMQListener(Component):
    """ A JMS listener for receiving the messages off WebSphere MQ queues.
    """
    def __init__(self):
        super(Component, self).__init__()
        self.logger = logging.getLogger("springpython.jms.listener.WebSphereMQListener(%s)" % (hex(id(self))))
        
    def _get_destination_info(self):
        return "destination=[%s], %s" % (self.destination, self.factory.get_connection_info())

    def run(self, *ignored):
        while True:
            try:
                message = self.factory.receive(self.destination, self.wait_interval)
                self.logger.log(TRACE1, "Message received [%s]" % str(message).decode("utf-8"))
                
                work_request = WorkRequest(self.handler.handle, [message])
                self.handlers_pool.putRequest(work_request)
                
                try:
                    self.handlers_pool.poll()
                except NoResultsPending, e:
                    pass

            except NoMessageAvailableException, e:
                self.logger.log(TRACE1, "Consumer did not receive a message. %s" % self._get_destination_info())
                
            except WebSphereMQJMSException, e:
                self.logger.error("%s in run, e.completion_code=[%s], "
                    "e.reason_code=[%s]" % (e.__class__.__name__, e.completion_code, e.reason_code))
                raise

class SimpleMessageListenerContainer(InitializingObject):
    """ A container for individual JMS listeners.
    """
    
    def __init__(self, factory=None, destination=None, handler=None, 
                 concurrent_listeners=1, handlers_per_listener=2, wait_interval=1000):
        """ factory - reference a to JMS connection factory
        destination - name of a queue to get the messages off
        handler - reference to an object which will be passed the incoming messages
        concurrent_listeners - how many concurrent JMS listeners the container
                               will manage
        handlers_per_listener - how many handler threads each listener will receive
        wait_interval - time, in milliseconds, indicating how often each JMS
                        listener will check for new messages
        """

        self.factory = factory
        self.destination = destination
        self.handler = handler
        self.concurrent_listeners = concurrent_listeners
        self.handlers_per_listener = handlers_per_listener
        self.wait_interval = wait_interval
        
        self.logger = logging.getLogger("springpython.jms.listener.SimpleMessageListenerContainer")
    
    def after_properties_set(self):
        """ Run by Spring Python after all the JMS container's properties have
        been set.
        """
        
        for idx in range(self.concurrent_listeners):
            # Create as many Circuits managers as there are JMS listeners.
            manager = Manager()
            manager.start()
            
            # A pool of handler threads for each listener.
            handlers_pool = ThreadPool(self.handlers_per_listener)
            
            # Each manager gets assigned its own listener.
            listener = WebSphereMQListener()
            
            # Assign the listener and a debugger component to the manager.
            manager += listener
            manager += Debugger(logger=self.logger)
            
            listener.factory = self.factory
            listener.destination = self.destination
            listener.handler = self.handler
            listener.handlers_pool = handlers_pool
            listener.wait_interval = self.wait_interval
            listener.start()



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
import logging
from springpython.security import AuthenticationException
from springpython.security import BadCredentialsException
from springpython.security import DisabledException

class ProviderNotFoundException(AuthenticationException):
    """
    An exception thrown when a list of providers are polled for a security decision,
    and none of them supports the request.
    """
    pass

class Authentication:
    """
    Abstract representation of credential data. The premise is that username and password
    are populated, and after authentication this record is returned with the third attribute,
    granted authorities, populated.
    """
    
    def __init__(self):
        self.__authenticated = False

    def isAuthenticated(self):
        return self.__authenticated

    def setAuthenticated(self, authenticated):
        self.__authenticated = authenticated
        
    def getCredentials(self):
        raise NotImplementedError()

    def __str__(self):
        raise AuthenticationException("You should be using a concrete authentication object")

class UsernamePasswordAuthenticationToken(Authentication):
    """
    A basic concrete version of authentication. Works for most scenarios.
    """
    
    def __init__(self, username = None, password = None, granted_auths = None):
        Authentication.__init__(self)
        self.username = username
        self.password = password
        if granted_auths is None:
            self.granted_auths = []
        else:
            self.granted_auths = granted_auths
            
    def getCredentials(self):
        return self.password

    def __str__(self):
        return "[UsernamePasswordAuthenticationToken] User: [%s] Password: [PROTECTED] GrantedAuthorities: %s Authenticated: %s" % \
            (self.username, self.granted_auths, self.isAuthenticated())

class AuthenticationManager:
    """
    Iterates an Authentication request through a list of AuthenticationProviders. 
    
    AuthenticationProviders are tried in order until one provides a non-null response.
    A non-null response indicates the provider had authority to decide on the authentication
    request and no further providers are tried. If an AuthenticationException is thrown by
    a provider, it is retained until subsequent providers are tried. If a subsequent provider
    successfully authenticates the request, the earlier authentication exception is disregarded
    and the successful authentication will be used. If no subsequent provider provides a
    non-null response, or a new AuthenticationException, the last AuthenticationException
    received will be used. If no provider returns a non-null response, or indicates it can
    even process an Authentication, the AuthenticationManager will throw a ProviderNotFoundException.
    """
    
    def __init__(self, auth_providers = None):
        if auth_providers is None:
            self.auth_providers = []
        else:
            self.auth_providers = auth_providers
        self.logger = logging.getLogger("springpython.security.providers.AuthenticationManager")

    def authenticate(self, authentication):
        """
        Attempts to authenticate the passed Authentication object, returning a fully
        populated Authentication object (including granted authorities) if successful.
        """
        authenticationException = ProviderNotFoundException()
        for auth_provider in self.auth_providers:
            try:
                results = auth_provider.authenticate(authentication)
                if results:
                    results.setAuthenticated(True)
                    return results
            except DisabledException, e: # Disabled means account found, but invalid
                raise e
            except AuthenticationException, e:
                authenticationException = e
        raise authenticationException

class AuthenticationProvider(object):
    """
    Indicates a class can process a specific Authentication implementation.
    """
    
    def authenticate(self, authentication):
        """
        Performs authentication with the same contract as AuthenticationManager.authenticate(Authentication).
        """
        raise NotImplementedError()
    
    def supports(self, authentication):
        """
        Returns true if this AuthenticationProvider supports the indicated Authentication object.
        """
        raise NotImplementedError()




"""
   Copyright 2006-2009 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""

import logging
import re
import sys
from springpython.security import AuthenticationException
from springpython.security import AuthenticationServiceException
from springpython.security import BadCredentialsException
from springpython.security import DisabledException
from springpython.security import UsernameNotFoundException
from springpython.security.providers import AuthenticationProvider
from springpython.security.providers import UsernamePasswordAuthenticationToken
from springpython.security.providers.dao import AbstractUserDetailsAuthenticationProvider
from springpython.security.providers.encoding import LdapShaPasswordEncoder

"""
The ldap library only works with CPython. You should NOT import this library directly.
"""
import ldap


class DefaultSpringSecurityContextSource(object):
    """
    This class is used to define the url of the ldap server. It expects a string like ldap://<server>:<port>/<baseDN>
    It provides functions to retrieve the parts
    """
    
    def __init__(self, url=None):
        self.url = url
        
    def server(self):
        """Extract the server's hostname/port from the url."""
        return self.url.split("ldap://")[1].split("/")[0].split(":")

    def base(self):
        """Extract the baseDN from the url."""
        return self.url.split("ldap://")[1].split("/")[1]

class BindAuthenticator(object):
    """
    This ldap authenticator uses binding to confirm the user's password. This means the password encoding
    depends on the ldap library's API as well as the directory server; NOT Spring Python's password
    hashing algorithms.
    """
    
    def __init__(self, context_source=None, user_dn_patterns="uid={0},ou=people"):
        self.context_source = context_source
        self.user_dn_patterns = user_dn_patterns
        self.logger = logging.getLogger("springpython.security.providers.Ldap.BindAuthenticator")

    def authenticate(self, authentication):
        """Using the user_dn_patterns, find the user's entry, and then bind to the entry using supplied credentials."""
        
        username = self.user_dn_patterns.replace("{0}", authentication.username)
        baseDn = self.context_source.base()

        parts = username.split(",")

        if len(parts) > 1:
            username = parts[0]
            baseDn = ",".join(parts[1:]) + "," + baseDn

        (host, port) = self.context_source.server()
        self.logger.debug("Opening connection to server %s/%s" % (host, int(port)))
        l = ldap.open(host, int(port))

        self.logger.debug("Searching for %s in %s" % (username, baseDn))
        result_set = l.search_s(baseDn, ldap.SCOPE_SUBTREE, username, None)

        if len(result_set) != 1:
            raise BadCredentialsException("Found %s entries at %s/%s. Should only be 1." % (len(result_set), baseDn, username))

        dn = result_set[0][0]
        self.logger.debug("Attempting to bind %s" % dn)
        try:
            l.simple_bind_s(dn, authentication.password)
            self.logger.debug("Successfully bound to server!")
            return (result_set[0],l)
        except Exception, e:
            self.logger.debug("Error %s" % e)
            raise BadCredentialsException("Invalid password")
        
class PasswordComparisonAuthenticator(object):
    """
    This ldap authenticator uses string comparison to confirm the user's password. This means a password encoder must
    be provided, or the default LdapShaPasswordEncoder will be used. It searched for the user's entry, fetches the
    password, and then does a string comparison to confirm the password.
    """

    def __init__(self, context_source=None, user_dn_patterns="uid={0},ou=people", password_attr_name="userPassword"):
        self.context_source = context_source
        self.user_dn_patterns = user_dn_patterns
        self.password_attr_name = password_attr_name
        self.encoder = LdapShaPasswordEncoder()
        self.logger = logging.getLogger("springpython.security.providers.Ldap.PasswordComparisonAuthenticator")

    def authenticate(self, authentication):
        """
        Using the user_dn_patterns, find the user's entry, and then retrieve the password field. Encode the supplied
        password with the necessary hasher, and compare to the entry.
        """

        username = self.user_dn_patterns.replace("{0}", authentication.username)
        baseDn = self.context_source.base()

        parts = username.split(",")

        if len(parts) > 1:
            username = parts[0]
            baseDn = ",".join(parts[1:]) + "," + baseDn

        (host, port) = self.context_source.server()
        self.logger.debug("Opening connection to server %s/%s" % (host, int(port)))
        l = ldap.open(host, int(port))

        self.logger.debug("Searching for %s in %s" % (username, baseDn))
        result_set = l.search_s(baseDn, ldap.SCOPE_SUBTREE, username, None)

        if len(result_set) != 1:
            raise BadCredentialsException("Found %s entries at %s/%s. Should only be 1." % (len(result_set), baseDn, username))

        self.logger.debug("Looking for attributes...%s" % result_set[0][1])
        stored_password = result_set[0][1][self.password_attr_name.lower()][0]
        self.logger.debug("Comparing passwords...")

        if self.encoder.isPasswordValid(stored_password, authentication.password, None):
            self.logger.debug("Successfully matched passwords!")
            return (result_set[0],l)
        else:
            raise BadCredentialsException("Invalid password")

class DefaultLdapAuthoritiesPopulator(object):
    """
    This ldap authorities populator follows a standard convention, where groups are created, with a member attribute, pointing
    at user entries in another part of the directory structure. It then combines ROLE_ with the name of the group, and names
    that as a granted role.
    """
    
    def __init__(self, context_source=None, group_search_base="ou=groups", group_search_filter="member={0}", group_role_attr="cn", role_prefix="ROLE_", convert_to_upper=True):
        self.logger = logging.getLogger("springpython.security.providers.Ldap.DefaultLdapAuthoritiesPopulator")
        self.context_source = context_source
        self.group_search_base = group_search_base
        self.group_search_filter = group_search_filter
        self.group_role_attr = group_role_attr 
        self.role_prefix = role_prefix
        self.convert_to_upper = convert_to_upper

    def get_granted_auths(self, user_details, l):
        group_filter = self.group_search_filter.replace("{0}", user_details[0])
        baseDn = self.group_search_base + "," + self.context_source.base()

        self.logger.debug("Searching for groups for %s" % str(user_details[0]))
        result_set = l.search_s(baseDn, ldap.SCOPE_SUBTREE, group_filter, None) 

        auths = []
        for row in result_set:
            role = self.role_prefix + row[1][self.group_role_attr][0]
            if self.convert_to_upper:
                auths.append(role.upper())
            else:
                auths.append(role)
        self.logger.debug("Authorities = %s" % auths)
        return auths
        
class LdapAuthenticationProvider(AuthenticationProvider):
    """
    This authenticator performs two steps:
        1) Authenticate the user to confirm their credentials.
        2) Lookup roles the user has stored in the directory server.
        
    It is possible to inject any type of authenticator as well as roles populator.
    
    Spring Python includes two authenticators that perform standard binding or password comparisons.
    You are able to code your own and use it instead, especially if you are using a non-conventional mechanism.
    
    Spring Python includes one role populator, based on the standard convention of defining groups elsewhere in
    the directory server's hierarchy. However, you can inject your own if you have a non-convential structure,
    such as storing the roles directly in the user's directory entry.
    """
    
    def __init__(self, ldap_authenticator=None, ldap_authorities_populator=None):
        AuthenticationProvider.__init__(self)
        self.ldap_authenticator = ldap_authenticator
        self.ldap_authorities_populator = ldap_authorities_populator
        self.logger = logging.getLogger("springpython.security.providers.Ldap.LdapAuthenticationProvider")
            
    def authenticate(self, authentication):
        user_details, l = self.ldap_authenticator.authenticate(authentication)
        from copy import deepcopy
        results = deepcopy(authentication)
        results.granted_auths = self.ldap_authorities_populator.get_granted_auths(user_details, l)
        l.unbind()
        return results
        



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
try:
    import cElementTree as etree
except ImportError:
    try:
        import xml.etree.ElementTree as etree
    except ImportError:
        from elementtree import ElementTree as etree

import re
import types
import inspect
import logging

from _config_base import *
from springpython.context import scope
from decorator import decorator, partial
from springpython.context import ApplicationContextAware
from springpython.factory import PythonObjectFactory
from springpython.factory import ReflectiveObjectFactory
from springpython.container import InvalidObjectScope

class PythonConfig(Config, ApplicationContextAware):
    """
    PythonConfig supports using pure python code to define objects.
    """

    def __init__(self):
        self.logger = logging.getLogger("springpython.config.PythonConfig")
        super(PythonConfig, self).__init__()

    def read_object_defs(self):
        self.logger.debug("==============================================================")
        objects = []
        self.logger.debug("Parsing %s" % self)
        for name, method in inspect.getmembers(self, inspect.ismethod):
            if name not in _pythonConfigMethods:
                try:
                    wrapper = method.im_func.func_globals["_call_"]

                    if wrapper.func_name.startswith("object"):
                        c = ObjectDef(id=name, factory=PythonObjectFactory(method, wrapper),
                                scope=wrapper.scope, lazy_init=wrapper.lazy_init,
                                abstract=wrapper.abstract, parent=wrapper.parent)
                        objects.append(c)
                except KeyError, e:
                    pass
        self.logger.debug("==============================================================")
        return objects

    def set_app_context(self, app_context):
        super(PythonConfig, self).set_app_context(app_context)
        try:
            _object_context[(self,)]["container"] = app_context
        except KeyError, e:
            _object_context[(self,)] = {"container": app_context}


_pythonConfigMethods = [name for (name, method) in inspect.getmembers(PythonConfig, inspect.ismethod)]

_object_context = {}

def _object_wrapper(f, scope, parent, log_func_name, *args, **kwargs):
    """
    This function checks if the object already exists in the container. If so,
    it will retrieve its results. Otherwise, it calls the function.

    For prototype objects, the function is basically a pass through,
    because everytime a prototype function is called, there should be no
    caching of results.

    Using the @decorator library greatly simplifies the implementation of this.
    """

    def _deco(f, scope, parent, log_func_name, *args, **kwargs):
        log = logging.getLogger("springpython.config.%s%s - %s%s" % (log_func_name,
                                    f, str(args), scope))
        if f.func_name != top_func:
            log.debug("This is NOT the top-level object %s, deferring to container." % top_func)
            container = _object_context[args]["container"]
            log.debug("Container = %s" % container)

            if parent:
                parent_result = container.get_object(parent, ignore_abstract=True)
                log.debug("This IS the top-level object, calling %s(%s)" \
                           % (f.func_name, parent_result))
                results = container.get_object(f.func_name)(parent_result)
            else:
                results = container.get_object(f.func_name)

            log.debug("Found %s inside the container" % results)
            return results
        else:
            if parent:
                container = _object_context[(args[0],)]["container"]
                parent_result = container.get_object(parent, ignore_abstract=True)
                log.debug("This IS the top-level object, calling %s(%s)" \
                           % (f.func_name, parent_result))
                results = f(container, parent_result)
            else:
                log.debug("This IS the top-level object, calling %s()." % f.func_name)
                results = f(*args, **kwargs)

            log.debug("Found %s" % results)
            return results

    return _deco(f, scope, parent, log_func_name, *args, **kwargs)

def Object(theScope=scope.SINGLETON, lazy_init=False, abstract=False, parent=None):
    """
    This function is a wrapper around the function which returns the real decorator.
    It decides, based on scope and lazy-init, which decorator to return.
    """
    if type(theScope) == types.FunctionType:
        return Object()(theScope)

    elif theScope == scope.SINGLETON:
        log_func_name = "objectSingleton"

    elif theScope == scope.PROTOTYPE:
        log_func_name = "objectPrototype"

    else:
        raise InvalidObjectScope("Don't know how to handle scope %s" % theScope)

    def object_wrapper(f, *args, **kwargs):
        return _object_wrapper(f, theScope, parent, log_func_name, *args, **kwargs)

    object_wrapper.lazy_init = lazy_init
    object_wrapper.abstract = abstract
    object_wrapper.parent = parent
    object_wrapper.scope = theScope

    return decorator(object_wrapper)



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
import logging
import re
from springpython.security import AuthenticationCredentialsNotFoundException
from springpython.security.context import SecurityContextHolder
from springpython.aop import MethodInterceptor
from springpython.aop import MethodInvocation

logger = logging.getLogger("springpython.security.intercept")

class ObjectDefinitionSource(object):
    """Implemented by classes that store and can identify the ConfigAttributeDefinition that applies to a given secure object invocation."""

    def get_attributes(obj):
        """Accesses the ConfigAttributeDefinition that applies to a given secure object."""
        raise NotImplementedError()

    def get_conf_attr_defs():
        """If available, all of the ConfigAttributeDefinitions defined by the implementing class."""
        raise NotImplementedError()

    def supports(cls):
        """Indicates whether the ObjectDefinitionSource implementation is able to provide ConfigAttributeDefinitions for
        the indicated secure object type."""
        raise NotImplementedError()

class InterceptorStatusToken(object):
    """
    A return object received by AbstractSecurityInterceptor subclasses.

    This class reflects the status of the security interception, so that the final call to
    AbstractSecurityInterceptor.after_invocation(InterceptorStatusToken, Object) can tidy up correctly.
    """
    
    def __init__(self, authentication = None, attr = None, secure_obj = None):
        self.authentication = authentication
        self.attr = attr
        self.secure_obj = secure_obj

class AbstractSecurityInterceptor(object):
    """
    Abstract class that implements security interception for secure objects.
    
    It will implements the proper handling of secure object invocations, being:
    
       1. Obtain the Authentication object from the SecurityContextHolder.
       2. Determine if the request relates to a secured or public invocation by looking up the secure object request
          against the ObjectDefinitionSource.
       3. For an invocation that is secured (there is a ConfigAttributeDefinition for the secure object invocation):
             1. If either the Authentication.isAuthenticated() returns false, or the alwaysReauthenticate is true,
                authenticate the request against the configured AuthenticationManager. When authenticated, replace
                the Authentication object on the SecurityContextHolder with the returned value.
             2. Authorize the request against the configured AccessDecisionManager.
             (3. Perform any run-as replacement via the configured RunAsManager. FUTURE)
             4. Pass control back to the concrete subclass, which will actually proceed with executing the object.
                An InterceptorStatusToken is returned so that after the subclass has finished proceeding with execution
                of the object, its finally clause can ensure the AbstractSecurityInterceptor is re-called and tidies up
                correctly.
             5. The concrete subclass will re-call the AbstractSecurityInterceptor via the after_invocation(InterceptorStatusToken, Object) method.
             (6. If the RunAsManager replaced the Authentication object, return the SecurityContextHolder to the object
                that existed after the call to AuthenticationManager. FUTURE)
             7. If an AfterInvocationManager is defined, invoke the invocation manager and allow it to replace the object
                due to be returned to the caller.
       (4. For an invocation that is public (there is no ConfigAttributeDefinition for the secure object invocation):
             1. As described above, the concrete subclass will be returned an InterceptorStatusToken which is subsequently
                re-presented to the AbstractSecurityInterceptor after the secure object has been executed. The
                AbstractSecurityInterceptor will take no further action when its after_invocation(InterceptorStatusToken, Object)
                is called. FUTURE)
       5. Control again returns to the concrete subclass, along with the Object that should be returned to the caller. The
          subclass will then return that result or exception to the original caller.
    """
    
    def __init__(self, auth_manager = None, access_decision_mgr = None, obj_def_source = None):
        self.auth_manager = auth_manager
        self.access_decision_mgr = access_decision_mgr
        self.obj_def_source = obj_def_source
        self.logger = logging.getLogger("springpython.security.intercept.AbstractSecurityInterceptor")

    def obtain_obj_def_source(self):
       raise NotImplementedError()

    def before_invocation(self, invocation):
        attr = self.obtain_obj_def_source().get_attributes(invocation)
        if attr:
            self.logger.debug("Secure object: %s; ConfigAttributes: %s" % (invocation, attr))
            if not SecurityContextHolder.getContext().authentication:
                raise AuthenticationCredentialsNotFoundException("An Authentication object was not found in the security credentials")
            if not SecurityContextHolder.getContext().authentication.isAuthenticated():
                authenticated = self.auth_manager.authenticate(SecurityContextHolder.getContext().authentication)
                self.logger.debug("Successfully Authenticated: " + authenticated)
                SecurityContextHolder.getContext().authentication = authenticated
            else:
                authenticated = SecurityContextHolder.getContext().authentication
                self.logger.debug("Previously Authenticated: %s" % authenticated)
            self.access_decision_mgr.decide(authenticated, invocation, attr)
            self.logger.debug("Authorization successful")
            return InterceptorStatusToken(authenticated, attr, invocation)
        else:
            return None
    
    def after_invocation(self, token, results):
        """As a minimum, this needs to pass the results right on through. Subclasses can extend this behavior
        to utilize the token information."""
        return results

class AbstractMethodDefinitionSource(ObjectDefinitionSource):
    """Abstract implementation of ObjectDefinitionSource."""
    
    def get_attributes(self, obj):
        try:
            module_name = obj.instance.__module__
            class_name = obj.instance.__class__.__name__
            method_name = obj.method_name
            full_method_name = "%s.%s.%s" % (module_name, class_name, method_name)
            return self.lookupAttributes(full_method_name)
        except AttributeError:
            raise TypeError("obj must be a MethodInvocation")

    def lookupAttributes(self, method):
        raise NotImplementedError()

class MethodDefinitionMap(AbstractMethodDefinitionSource):
    """
    Stores an obj_def_source for each method signature defined in a object.
    
    Regular expressions are used to match a method request in a ConfigAttributeDefinition. The order of registering
    the regular expressions is very important. The system will identify the first matching regular expression for a given
    method. It will not proceed to evaluate later regular expressions if a match has already been found.
    
    Accordingly, the most specific regular expressions should be registered first, with the most general regular expressions registered last.    
    """
    
    def __init__(self, obj_def_source):
        self.obj_def_source = obj_def_source

    def lookupAttributes(self, method):
        if self.obj_def_source:
            for rule, attr in self.obj_def_source:
                if re.compile(rule).match(method):
                    return attr 
        return None

class MethodSecurityInterceptor(MethodInterceptor, AbstractSecurityInterceptor):
    """
    Provides security interception of Spring Python AOP-based method invocations.

    The ObjectDefinitionSource required by this security interceptor is of type MethodDefinitionMap.

    Refer to AbstractSecurityInterceptor for details on the workflow.   
    """
    
    def __init__(self):
        MethodInterceptor.__init__(self)
        AbstractSecurityInterceptor.__init__(self)
        self.validate_config_attributes = False
        self.obj_def_source = None

    def __setattr__(self, name, value):
        if name == "obj_def_source" and value is not None:
            self.__dict__[name] = MethodDefinitionMap(value)
        else:
            self.__dict__[name] = value

    def obtain_obj_def_source(self):
        return self.obj_def_source

    def invoke(self, invocation):
        token = self.before_invocation(invocation)
        results = None
        try:
            results = invocation.proceed()
        finally:
            results = self.after_invocation(token, results)
        return results



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import copy
import logging
import re
import types
from springpython.aop import utils

class Pointcut(object):
    """Interface defining where to apply an aspect."""
    def class_filter(self):
        raise NotImplementedError()
    def method_matcher(self):
        raise NotImplementedError()

class MethodMatcher(object):
    """Interface defining how to apply aspects based on methods."""
    def matches_method_and_target(self, method, targetClass, args):
        raise NotImplementedError()

class MethodInterceptor(object):
    """Interface defining "around" advice."""
    def invoke(self, invocation):
        raise NotImplementedError()

class MethodInvocation(object):
    """Encapsulation of invoking a method on a proxied service. It iterates throgh the list of interceptors by using
    a generator."""
    def __init__(self, instance, method_name, args, kwargs, interceptors):
        self.instance = instance
        self.method_name = method_name
        self.args = args
        self.kwargs = kwargs
        self.intercept_stack = copy.copy(interceptors)
        self.intercept_stack.append(FinalInterceptor())
        self.logger = logging.getLogger("springpython.aop.MethodInvocation")

    def getInterceptor(self):
        """This is a generator to proceed through the stack of interceptors. By using generator convention, code may
        proceed in a nested fashion, versus a for-loop which would act in a chained fashion."""
        for interceptor in self.intercept_stack:
            yield interceptor

    def proceed(self):
        """This is the method every interceptor should call in order to continue down the chain of interceptors."""
        interceptor = self.iterator.next()
        self.logger.debug("Calling %s.%s(%s, %s)" % (interceptor.__class__.__name__, self.method_name, self.args, self.kwargs))
        return interceptor.invoke(self)

    def __getattr__(self, name):
        """This only deals with method invocations. Attributes are dealt with by the AopProxy, and don't every reach this
        block of code."""
        self.iterator = self.getInterceptor()
        self.method_name = name
        return self

    def __call__ (self, *args, **kwargs):
        """This method converts this from being a stored object into a callable class. This is effectively like a metaclass
        that dispatches calls to proceed through a stack of interceptors."""
        self.args = args
        self.kwargs = kwargs
        return self.proceed()

    def dump_interceptors(self, level = logging.INFO):
        """DEBUG: Method used to dump the stack of interceptors in order of execution."""
        for interceptor in self.intercept_stack:
            self.logger.log(level, "Interceptor stack: %s" % interceptor.__class__.__name__)

class RegexpMethodPointcutAdvisor(Pointcut, MethodMatcher, MethodInterceptor):
    """
    This is a combination PointCut/MethodMatcher/MethodInterceptor. It allows associating one or more
    defined advices with a set of regular expression patterns.
    """
    def __init__(self, advice = None, patterns = None):
        Pointcut.__init__(self)
        MethodMatcher.__init__(self)
        self.advice = advice
        if not patterns:
            self.patterns = []
        else:
            self.patterns = patterns
        self.logger = logging.getLogger("springpython.aop.RegexpMethodPointcut")

    def init_patterns(self):
        """Precompile the regular expression pattern matcher list."""
        self.compiled_patterns = {}
        for pattern in self.patterns:
            self.compiled_patterns[pattern] = re.compile(pattern)

    def matches_method_and_target(self, method, target_class, args):
        """Iterate through all patterns, checking for a match. Calls the pattern matcher against "class.method_name"."""
        for pointcut_pattern in self.patterns:
            if (self.matches_pattern(target_class + "." + method, pointcut_pattern)):
                return True
        return False

    def matches_pattern(self, method_name, pointcut_pattern):
        """Uses a pre-built dictionary of regular expression patterns to check for a matcch."""
        if self.compiled_patterns[pointcut_pattern].match(method_name):
            matched = True
        else:
            matched = False
        self.logger.debug("Candidate is [%s]; pattern is [%s]; matched=%s" % (method_name, pointcut_pattern, matched))
        return matched

    def invoke(self, invocation):
        """Compares "class.method" against regular expression pattern and if it passes, it will
        pass through to the chain of interceptors. Otherwise, bypass interceptors and invoke class
        method directly."""

        className = invocation.instance.__class__.__name__

        if self.matches_method_and_target(invocation.method_name, className, invocation.args):
            # This constant is not class level, because it is a hack solution, and should only be used
            # used here, and not known outside the scope of this block of code. --G.Turnquist (9/22/2008)
            ASSUME_THIS_ADVISOR_WAS_FIRST = 1
            invocation.intercept_stack[ASSUME_THIS_ADVISOR_WAS_FIRST:ASSUME_THIS_ADVISOR_WAS_FIRST] = self.advice

            self.logger.debug("We have a match, passing through to the advice.")
            invocation.dump_interceptors(logging.DEBUG)

            return invocation.proceed()
        else:
            self.logger.debug("No match, bypassing advice, going straight to targetClass.")
            return getattr(invocation.instance, invocation.method_name)(*invocation.args, **invocation.kwargs)

    def __setattr__(self, name, value):
        """If "advice", make sure it is a list. Anything else, pass through to simple assignment.
        Also, if "patterns", initialize the regular expression parsers.
        """
        if name == "advice" and type(value) != list:
            self.__dict__[name] = [value]
        else:
            self.__dict__[name] = value

        if name == "patterns":
            self.init_patterns()

class FinalInterceptor(MethodInterceptor):
    """
    Final interceptor is always at the bottom of interceptor stack.
    It executes the actual target method on the instance.
    """
    def __init__(self):
        MethodInterceptor.__init__(self)
        self.logger = logging.getLogger("springpython.aop.FinalInterceptor")

    def invoke(self, invocation):
        return getattr(invocation.instance, invocation.method_name)(*invocation.args, **invocation.kwargs)

class AopProxy(object):
    """AopProxy acts like the target object by dispatching all method calls to the target through a MethodInvocation.
    The MethodInvocation object actually deals with potential "around" advice, referred to as interceptors. Attribute
    lookups are not intercepted, but instead fetched from the actual target object."""

    def __init__(self, target, interceptors):
        self.target = target
        if type(interceptors) == list:
            self.interceptors = interceptors
        else:
            self.interceptors = [interceptors]
        self.logger = logging.getLogger("springpython.aop.AopProxy")

    def __getattr__(self, name):
        """If any of the parameters are local objects, they are immediately retrieved. Callables cause the dispatch method
        to be return, which forwards callables through the interceptor stack. Target attributes are retrieved directly from
        the target object."""
        if name in ["target", "interceptors", "method_name"]:
            return self.__dict__[name]
        else:
            attr = getattr(self.target, name)
            if not callable(attr):
                return attr

            def dispatch(*args, **kwargs):
                """This method is returned to the caller emulating the function call being sent to the
                target object. This services as a proxying agent for the target object."""
                invocation = MethodInvocation(self.target, name, args, kwargs, self.interceptors)
                ##############################################################################
                # NOTE:
                # getattr(invocation, name) doesn't work here, because __str__ will print
                # the MethodInvocation's string, instead of trigger the interceptor stack.
                ##############################################################################
                return invocation.__getattr__(name)(*args, **kwargs)

            return dispatch

class ProxyFactory(object):
    """This object helps to build AopProxy objects programmatically. It allows configuring advice and target objects.
    Then it will produce an AopProxy when needed. To use similar behavior in an IoC environment, see ProxyFactoryObject."""

    def __init__(self, target = None, interceptors = None):
        self.logger = logging.getLogger("springpython.aop.ProxyFactory")
        self.target = target
        if not interceptors:
            self.interceptors = []
        elif type(interceptors) == list:
            self.interceptors = interceptors
        else:
            self.interceptors = [interceptors]

    def getProxy(self):
        """Generate an AopProxy given the current target and list of interceptors. Any changes to the factory after
        proxy creation do NOT propagate to the proxies."""
        return AopProxy(self.target, self.interceptors)

    def __setattr__(self, name, value):
        if name == "target" and type(value) == types.StringType:
            value = utils.getClass(value)()
        elif name == "interceptors" and not isinstance(value, list):
            value = [value]

        self.__dict__[name] = value

class ProxyFactoryObject(ProxyFactory, AopProxy):
    """This class acts as both a ProxyFactory to build and an AopProxy. It makes itself look like the target object.
    Any changes to the target and list of interceptors is immediately seen when using this as a proxy."""
    def __init__(self, target = None, interceptors = None):
        ProxyFactory.__init__(self, target, interceptors)
        self.logger = logging.getLogger("springpython.aop.ProxyFactoryObject")

    def __str__(self):
        return self.__getattr__("__str__")()

class PerformanceMonitorInterceptor(MethodInterceptor):
    def __init__(self, prefix = None, level = logging.DEBUG):
        self.prefix = prefix
        self.level = level
        self.logger = logging.getLogger("springpython.aop")

    def invoke(self, invocation):
        self.logger.log(self.level, "%s BEGIN" % (self.prefix))
        timing.start()
        results = invocation.proceed()
        timing.finish()
        self.logger.log(self.level, "%s END => %s" % (self.prefix, timing.milli()/1000.0))
        return results



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import logging
import traceback
from threading import RLock, currentThread

try:
    from cStringIO import StringIO
except ImportError, e:
    from StringIO import StringIO


TRACE1 = 6
logging.addLevelName(TRACE1, "TRACE1")

# Original code by Anand Balachandran Pillai (abpillai at gmail.com)
# http://code.activestate.com/recipes/533135/
class synchronized(object):
    """ Class enapsulating a lock and a function allowing it to be used as
    a synchronizing decorator making the wrapped function thread-safe """

    def __init__(self, *args):
        self.lock = RLock()
        self.logger = logging.getLogger("springpython.util.synchronized")

    def __call__(self, f):
        def lockedfunc(*args, **kwargs):
            try:
                self.lock.acquire()
                self.logger.log(TRACE1, "Acquired lock [%s] thread [%s]" % (self.lock, currentThread()))
                try:
                    return f(*args, **kwargs)
                except Exception, e:
                    raise
            finally:
                self.lock.release()
                self.logger.log(TRACE1, "Released lock [%s] thread [%s]" % (self.lock, currentThread()))
        return lockedfunc



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
import logging
from springpython.security import UsernameNotFoundException

class User(object):
    """
    Models core user information retieved by an UserDetailsService.
    """
    
    def __init__(self, username, password, enabled, accountNonExpired=True, accountNonLocked=True, credentialsNonExpired=True, authorities=None):
        self.username = username
        self.password = password
        if authorities is None:
            self.authorities = []
        else:
            self.authorities = authorities
        self.accountNonExpired = accountNonExpired
        self.accountNonLocked = accountNonLocked
        self.credentialsNonExpired = credentialsNonExpired
        self.enabled = enabled
        
    def __str__(self):
        return "Username: %s Password: [PROTECTED] Authorities: %s Enabled: %s" % (self.username, self.authorities, self.enabled)
        
class UserDetailsService(object):
    """
    Defines an interface for implementations that wish to provide data access services to the DaoAuthenticationProvider.

    The interface requires only one read-only method, which simplifies support of new data access strategies.
    """

    def load_user(self, username):
        raise NotImplementedError()

class InMemoryUserDetailsService(UserDetailsService):
    def __init__(self, user_dict = None):
        super(InMemoryUserDetailsService, self).__init__()
        if user_dict is None:
            self.user_dict = {}
        else:
            self.user_dict = user_dict
        self.logger = logging.getLogger("springpython.security.userdetails.InMemoryUserDetailsService")
        
    def load_user(self, username):
        if username in self.user_dict and len(self.user_dict[username][1]) > 0:
            self.logger.debug("Found %s in %s" % (username, self.user_dict))
            return User(username, self.user_dict[username][0], self.user_dict[username][2], True, True, True, self.user_dict[username][1])
        
        error = None
        if username not in self.user_dict:
            error = UsernameNotFoundException("User not found in %s" % self.user_dict)
        else:
            error = UsernameNotFoundException("User has no GrantedAuthority")
        self.logger.debug(error)
        raise error



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
import logging
from springpython.database.core import DatabaseTemplate
from springpython.database.core import RowMapper
from springpython.security import UsernameNotFoundException
from springpython.security.userdetails import User
from springpython.security.userdetails import UserDetailsService

class DatabaseUserDetailsService(UserDetailsService):
    """
    Retrieves user details (username, password, enabled flag, and authorities) from a database location.
 
    A default database structure is assumed, (see DEF_USERS_BY_USERNAME_QUERY and DEF_AUTHORITIES_BY_USERNAME_QUERY,
    which most users of this class will need to override, if using an existing scheme. This may be done by
    setting the default query strings used. If this does not provide enough flexibility, another strategy
    would be to subclass this class and override relevant parts.

    In order to minimise backward compatibility issues, this DAO does not recognise the expiration of user
    accounts or the expiration of user credentials. However, it does recognise and honour the user enabled/disabled
    column.   
    """
    
    DEF_USERS_BY_USERNAME_QUERY = "SELECT username,password,enabled FROM users WHERE username = ?"
    DEF_AUTHORITIES_BY_USERNAME_QUERY = "SELECT username,authority FROM authorities WHERE username = ?"

    class UsersByUsernameMapping(RowMapper):
        """A row handler that processes one user entry."""
        def map_row(self, row, metadata=None):
            username = row[0]
            password = row[1]
            enabled = row[2]
            return User(username, password, enabled, True, True, True, ["HOLDER"])
    
    class AuthoritiesByUsernameMapping(RowMapper):
        """A row handler that processes one granted authority for a given user."""
        def __init__(self, role_prefix):
            self.role_prefix = role_prefix
            
        def map_row(self, row, metadata=None):
            return self.role_prefix + row[1]
    
    def __init__(self, dataSource = None):
        super(DatabaseUserDetailsService, self).__init__()
        self.users_by_username_query = self.DEF_USERS_BY_USERNAME_QUERY
        self.auth_by_username_query = self.DEF_AUTHORITIES_BY_USERNAME_QUERY
        self.dataSource = dataSource
        self.role_prefix = ""
        self.username_based_pk = True
        self.logger = logging.getLogger("springpython.security.userdetails.DatabaseUserDetailsService")
        
    def load_user(self, username):
        dt = DatabaseTemplate(self.dataSource)
        
        users = dt.query(self.users_by_username_query, (username,), self.UsersByUsernameMapping())

        if len(users) == 0:
            raise UsernameNotFoundException("User not found")

        user = users[0] # First item in list, first column of tuple, containing no GrantedAuthority[]
        dbAuths = dt.query(self.auth_by_username_query, (user.username,), self.AuthoritiesByUsernameMapping(self.role_prefix))
        self.add_custom_authorities(user.username, dbAuths)

        if len(dbAuths) == 0:
            raise UsernameNotFoundException("User has no GrantedAuthority")

        auths = [dbAuth for dbAuth in dbAuths]
        return_username = user.username

        if not self.username_based_pk:
            return_username = username
            
        self.logger.debug("Just fetched %s from the database" % user)
        return User(return_username, user.password, user.enabled, True, True, True, auths)
    
    def add_custom_authorities(self, username, authorities):
        pass



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
import inspect
import logging
import re
import types
from springpython.aop import MethodInterceptor
from springpython.aop import ProxyFactoryObject
from springpython.context import ObjectPostProcessor
from springpython.config.decorator import decorator

logger = logging.getLogger("springpython.database.transaction")

class TransactionException(Exception):
    pass

class TransactionPropagationException(TransactionException):
    pass

class TransactionStatus(object):
    pass

class DefaultTransactionStatus(TransactionStatus):
    pass

class PlatformTransactionManager(object):
    """This interface is used to define the operations necessary in handling transactions."""
    def commit(self, status):
        raise NotImplementedError()

    def getTransaction(self, definition):
        raise NotImplementedError()

    def rollback(self, status):
        raise NotImplementedError()

class ConnectionFactoryTransactionManager(PlatformTransactionManager):
    """
    This transaction manager is based upon using a connection factory to control transactions. Since
    connection factories are tied to vendor-specific databases, this allows delegation of various
    transactional functions on a per-vendor basis.
    """

    def __init__(self, connection_factory):
        self.connection_factory = connection_factory
        self.logger = logging.getLogger("springpython.database.transaction.ConnectionFactoryTransactionManager")
        self.status = []

    def getTransaction(self, definition):
        """According to PEP 249, commits and rollbacks silently start new transactions. Until a more
        robust transaction manager is implemented to handle save points and so forth, this must suffice."""

        self.logger.debug("Analyzing %s" % definition.propagation)

        start_tx = False

        if definition.propagation == "PROPAGATION_REQUIRED":
            if len(self.status) == 0:
                self.logger.debug("There is no current transaction, and one is required, so starting one.")
                start_tx = True
            self.status.append(DefaultTransactionStatus())

        elif definition.propagation == "PROPAGATION_SUPPORTS":
            self.logger.debug("This code can execute inside or outside a transaction.")

        elif definition.propagation == "PROPAGATION_MANDATORY":
            if len(self.status) == 0:
                raise TransactionPropagationException("Trying to execute PROPAGATION_MANDATORY operation while outside TX")
            self.status.append(DefaultTransactionStatus())

        elif definition.propagation == "PROPAGATION_NEVER":
            if len(self.status) != 0:
                raise TransactionPropagationException("Trying to execute PROPAGATION_NEVER operation while inside TX")

        else:
            raise TransactionPropagationException("Transaction propagation level %s is not supported!" % definition.start_tx)

        if start_tx:
            self.logger.debug("START TRANSACTION")
            self.logger.debug("Creating a transaction, propagation = %s, isolation = %s, timeout = %s, read_only = %s" % (definition.propagation, definition.isolation, definition.timeout, definition.read_only))
            self.connection_factory.commit()

        return self.status

    def commit(self, status):
        self.status = status
        try:
            self.status.pop()
            if len(self.status) == 0:
                self.logger.debug("Commit the changes")
                self.connection_factory.commit()
                self.logger.debug("END TRANSACTION")
        except IndexError:
            pass

    def rollback(self, status):
        self.status = status
        try:
            self.status.pop()
            if len(self.status) == 0:
                self.logger.debug("Rolling back the transaction.")
                self.connection_factory.rollback()
                self.logger.debug("END TRANSACTION")
        except IndexError:
            pass

class TransactionDefinition(object):
    def __init__(self, isolation = None, name = None, propagation = None, timeout = None, read_only = None):
        self.isolation = isolation
        self.name = name
        self.propagation = propagation
        self.timeout = timeout
        self.read_only = read_only

class DefaultTransactionDefinition(TransactionDefinition):
    def __init__(self, isolation = "ISOLATION_DEFAULT", name = "", propagation = "PROPAGATION_REQUIRED", timeout = "TIMEOUT_DEFAULT", read_only = False):
        TransactionDefinition.__init__(self, isolation, name, propagation, timeout, read_only)

class TransactionTemplate(DefaultTransactionDefinition):
    """This utility class is used to simplify defining transactional blocks. Any exceptions thrown inside the
    transaction block will be propagated to whom ever is calling the template execute method."""

    def __init__(self, tx_manager):
        DefaultTransactionDefinition.__init__(self)	
        self.tx_manager = tx_manager
        self.logger = logging.getLogger("springpython.database.transaction.TransactionTemplate")

    def execute(self, transactionCallback):
        """Execute the action specified by the given callback object within a transaction."""

        status = self.tx_manager.getTransaction(self)
        result = None
        try:
            self.logger.debug("Execute the steps inside the transaction")
            result = transactionCallback.do_in_transaction(status)
            self.tx_manager.commit(status)
        except Exception, e:
            self.logger.debug("Exception: (%s)" % e)
            self.tx_manager.rollback(status)
            raise e
        return result

    def setTxAttributes(self, tx_attributes):
        for tx_def_prop in tx_attributes:
            if tx_def_prop.startswith("ISOLATION"):
                if tx_def_prop != self.isolation:       self.isolation = tx_def_prop
            elif tx_def_prop.startswith("PROPAGATION"):
                if tx_def_prop != self.propagation:     self.propagation = tx_def_prop
            elif tx_def_prop.startswith("TIMEOUT"):
                if tx_def_prop != self.timeout:         self.timeout = tx_def_prop
            elif tx_def_prop == "read_only":
                if not self.read_only:                  self.read_only = True
            else:
                self.logger.debug("Don't know how to handle %s" % tx_def_prop)


class TransactionCallback(object):
    """This interface defines the basic action needed to plug into the TransactionTemplate"""
    def do_in_transaction(self, status):
        raise NotImplementedError()

class TransactionCallbackWithoutResult(TransactionCallback):
    """This abstract class implements the TransactionCallback, but assumes no value is being returned."""
    def __init__(self):
        self.logger = logging.getLogger("springpython.database.transaction.TransactionCallbackWithoutResult")

    def do_in_transaction(self, status):
        self.logger.debug("Starting a transaction without result")
        self.do_in_tx_without_result(status)
        self.logger.debug("Completing a transaction without result")
        return None

    def do_in_tx_without_result(self, status):
        pass

class TransactionalInterceptor(MethodInterceptor):
    """This interceptor is used by the TransactionProxyFactoryObject in order to wrap
    method calls with transactions."""
    def __init__(self, tx_manager, tx_attributes):
        self.logger = logging.getLogger("springpython.database.transaction.TransactionalInterceptor")
        self.tx_attributes = tx_attributes
        self.tx_manager = tx_manager

    def invoke(self, invocation):
        class tx_def(TransactionCallback):
            def do_in_transaction(s, status):
                return invocation.proceed()

        tx_template = TransactionTemplate(self.tx_manager)

        # Iterate over the tx_attributes, and when a method match is found, apply the properties
        for pattern, tx_def_props in self.tx_attributes:
            if re.compile(pattern).match(invocation.method_name):
                self.logger.debug("%s matches pattern %s, tx attributes = %s" % (invocation.method_name, pattern, tx_def_props))
                tx_template.setTxAttributes(tx_def_props)
                break
            
        self.logger.debug("Call TransactionTemplate")
        try:
            results = tx_template.execute(tx_def())
        except Exception, e:
            self.logger.debug("Exception => %s" % e)
            raise e
        self.logger.debug("Return from TransactionTemplate")
        return results

class TransactionProxyFactoryObject(ProxyFactoryObject):
    """This class acts like the target object, and routes function calls through a
    transactional interceptor."""
    def __init__(self, tx_manager, target, tx_attributes):
        self.logger = logging.getLogger("springpython.database.transaction.TransactionProxyFactoryObject")
        ProxyFactoryObject.__init__(self, target, TransactionalInterceptor(tx_manager, tx_attributes))

def transactional(tx_attributes = None):
    """
    This decorator is actually a utility function that returns an embedded decorator, in order
    to handle whether it was called in any of the following ways:

    @transactional()
    def foo():
        pass

    @transactional
    def foo():
        pass

    The first two ways get parsed by Python as:

    foo = transactional("some contextual string")(foo)      # first way
    foo = transactional()(foo)                              # second way

    Since this is expected, they are granted direct access to the embedded transactional_wrapper.

    However, the third way ends up getting parsed by Python as:

    foo = Transactional(foo)

    This causes context to improperly get populated with a function instead of a string. This
    requires recalling this utility like:

    return Transactional()(context)
    """

    @decorator
    def transactional_wrapper(f, *args, **kwargs):
        """
        transactional_wrapper is used to wrap the decorated function in a TransactionTemplate callback,
        and then return the results.
        """
        class tx_def(TransactionCallback):
            """TransactionTemplate requires a callback defined this way."""
            def do_in_transaction(s, status):
                return f(*args, **kwargs)

        try:
            # Assumes tx_manager is supplied by AutoTransactionalObject
            tx_template = TransactionTemplate(tx_manager)
            if tx_attributes is not None:
                tx_template.setTxAttributes(tx_attributes)
            else:
                logger.debug("There are NO tx_attributes! %s" % tx_attributes)
            return tx_template.execute(tx_def())
        except NameError:
            # If no AutoTransactionalObject found in IoC container, then pass straight through.
            return tx_def().do_in_transaction(None)

    if type(tx_attributes) == types.FunctionType:
        return transactional()(tx_attributes)
    else:
        return transactional_wrapper


class AutoTransactionalObject(ObjectPostProcessor):
    """
    This object is used to automatically scan objects in an IoC container, and if @Transaction
    is found applied to any of the object's methods, link it with a TransactionManager.
    """

    def __init__(self, tx_manager):
        self.tx_manager = tx_manager
        self.logger = logging.getLogger("springpython.database.transaction.AutoTransactionalObject")

    def post_process_after_initialization(self, obj, obj_name):
        """This setup is run after all objects in the container have been created."""
        # Check every method in the object...
        for name, method in inspect.getmembers(obj, inspect.ismethod):
            try:
                # If the method contains _call_, then you are looking at a wrapper...
                wrapper = method.im_func.func_globals["_call_"]
                if wrapper.func_name == "transactional_wrapper":  # name of @transactional's wrapper method
                    self.logger.debug("Linking tx_manager with %s" % name)
                    wrapper.func_globals["tx_manager"] = self.tx_manager
            except KeyError, e:   # If the method is NOT wrapped, there will be no _call_ attribute
                pass
        return obj





"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import cherrypy
import logging

from springpython.security import AuthenticationException
from springpython.security.context import SecurityContextHolder, SecurityContext
from springpython.security.intercept import AbstractSecurityInterceptor
from springpython.security.providers import UsernamePasswordAuthenticationToken
from springpython.security.web import FilterChainProxy, SessionStrategy,RegExpBasedFilterInvocationDefinitionMap, FilterInvocation

class CP3FilterChainProxy(FilterChainProxy):
    def __init__(self, filterInvocationDefinitionSource=None):
        FilterChainProxy.__init__(self, filterInvocationDefinitionSource)
        self.logger = logging.getLogger("springpython.security.cherrypy3.Another")
        cherrypy.tools.securityFilterChain = cherrypy._cptools.HandlerTool(self)

    def __call__(self, environ=None, start_response=None):
        if cherrypy.request.handler is None:
            return False

        def cherrypy_handler(*args, **kwargs):
            return cherrypy.request.handler(*args, **kwargs)

        def func(args):
           return args[0]()

        self.application = (func, (cherrypy_handler,))
        cherrypy.response.body = FilterChainProxy.__call__(self, cherrypy.request.wsgi_environ, None)
        return True

class CP3SessionStrategy(SessionStrategy):
    def __init__(self):
        SessionStrategy.__init__(self)
        self.logger = logging.getLogger("springpython.security.cherrypy3.CP3SessionStrategy")

    def getHttpSession(self, environ):
        return cherrypy.session.get("session_id")

    def setHttpSession(self, key, value):
        if "session_id" not in cherrypy.session:
            cherrypy.session["session_id"] = {}
        cherrypy.session["session_id"][key] = value

class CP3RedirectStrategy(object):
    def redirect(self, url):
        raise cherrypy.HTTPRedirect(url)



"""
   Copyright 2006-2009 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
import logging
import re
import sys
from springpython.security import AuthenticationException
from springpython.security import AuthenticationServiceException
from springpython.security import BadCredentialsException
from springpython.security import DisabledException
from springpython.security import UsernameNotFoundException
from springpython.security.providers import AuthenticationProvider
from springpython.security.providers import UsernamePasswordAuthenticationToken
from springpython.security.providers.dao import AbstractUserDetailsAuthenticationProvider
from springpython.security.providers.encoding import LdapShaPasswordEncoder


"""
The ldap library only works with Jython. You should NOT import this library directly.

Due to the lack of a pure Python library, this version uses Spring Security/Spring LDAP jar files to perform
authentication and LDAP lookups.
"""
import java
import org.springframework.security.ldap.DefaultSpringSecurityContextSource
import org.springframework.security.ldap.populator.DefaultLdapAuthoritiesPopulator
import org.springframework.security.providers.ldap.authenticator.BindAuthenticator
import org.springframework.security.providers.ldap.authenticator.PasswordComparisonAuthenticator
import org.springframework.security.providers.UsernamePasswordAuthenticationToken
from jarray import array

print """
WARNING WARNING WARNING WARNING
===============================
This doesn't yet work. There is some issue with Jython.
See http://bugs.jython.org/issue1489 and http://jira.springframework.org/browse/SESPRINGPYTHONPY-121 for more details.
===============================
WARNING WARNING WARNING WARNING
"""

class DefaultSpringSecurityContextSource(object):
    def __init__(self, url):
        self._context = org.springframework.security.ldap.DefaultSpringSecurityContextSource(url)
        java.lang.Thread.currentThread().setContextClassLoader(self._context.getClass().getClassLoader())
        self._context.afterPropertiesSet()
        
class BindAuthenticator(object):
    def __init__(self, context_source=None, user_dn_patterns="uid={0},ou=people"):
        self.context_source = context_source
        self.user_dn_patterns = user_dn_patterns
        self.logger = logging.getLogger("springpython.security.providers.Ldap.BindAuthenticator")
        self._authenticator = None

    def authenticate(self, authentication):
        if self._authenticator is None:
            self._authenticator = org.springframework.security.providers.ldap.authenticator.BindAuthenticator(self.context_source._context)
            self._authenticator.setUserDnPatterns(array([self.user_dn_patterns], java.lang.String))
            self._authenticator.afterPropertiesSet()
            #java.lang.Thread.currentThread().setContextClassLoader(self._authenticator.getClass().getClassLoader())
            #print "BindAuthenticator class loader %s" % self._authenticator.getClass().getClassLoader()
        token = org.springframework.security.providers.UsernamePasswordAuthenticationToken(authentication.username, authentication.password)
        return self._authenticator.authenticate(token)
        
class PasswordComparisonAuthenticator(object):
    def __init__(self, context_source=None, user_dn_patterns="uid={0},ou=people", password_attr_name="userPassword"):
        self.context_source = context_source
        self.user_dn_patterns = user_dn_patterns
        self.password_attr_name = password_attr_name
        self.encoder = LdapShaPasswordEncoder()
        self.logger = logging.getLogger("springpython.security.providers.Ldap.PasswordComparisonAuthenticator")

    def authenticate(self, authentication):
        if jython:
            raise Exception("This code doesn't work inside Jython.")

class DefaultLdapAuthoritiesPopulator(object):
    def __init__(self, context_source=None, group_search_base="ou=groups", group_search_filter="(member={0})", group_role_attr="cn", role_prefix="ROLE_", convert_to_upper=True):
        self.logger = logging.getLogger("springpython.security.providers.Ldap.DefaultLdapAuthoritiesPopulator")
        self.context_source = context_source
        self.group_search_base = group_search_base
        self.group_search_filter = group_search_filter
        self.group_role_attr = group_role_attr 
        self.role_prefix = role_prefix
        self.convert_to_upper = convert_to_upper
        self._populator = org.springframework.security.ldap.populator.DefaultLdapAuthoritiesPopulator(self.context_source._context, self.group_search_base)
        #java.lang.Thread.currentThread().setContextClassLoader(self._populator.getClass().getClassLoader())
        self._populator.setGroupSearchFilter(self.group_search_filter)
        self._populator.setGroupRoleAttribute(self.group_role_attr)
        self._populator.setRolePrefix(self.role_prefix)
        self._populator.setConvertToUpperCase(self.convert_to_upper)
        print "LdapAuthoritiesPopulator class loader %s" % self._populator.getClass().getClassLoader()

    def get_granted_auths(self, user_details, username):
        results = self._populator.getGrantedAuthorities(user_details, username)
        print results
        return results
        
class LdapAuthenticationProvider(AuthenticationProvider):
    def __init__(self, ldap_authenticator=None, ldap_authorities_populator=None):
        AuthenticationProvider.__init__(self)
        self.ldap_authenticator = ldap_authenticator
        self.ldap_authorities_populator = ldap_authorities_populator
        self.logger = logging.getLogger("springpython.security.providers.Ldap.LdapAuthenticationProvider")
            
    def authenticate(self, authentication):
        user_details = self.ldap_authenticator.authenticate(authentication)
        print "Context class loader %s" % user_details.getClass().getClassLoader()
        from copy import deepcopy
        results = deepcopy(authentication)
        results.granted_auths = self.ldap_authorities_populator.get_granted_auths(user_details, authentication.username)
        results.setAuthenticated(True)
        l.unbind()
        return results
        



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

# stdlib
import sys
import logging
from threading import RLock
from cStringIO import StringIO
from struct import pack, unpack
from xml.sax.saxutils import escape
from binascii import hexlify, unhexlify
from time import time, mktime, strptime, altzone
from traceback import format_exc

try:
    import cElementTree as etree
except ImportError:
    try:
        import xml.etree.ElementTree as etree
    except ImportError:
        from elementtree import ElementTree as etree

# Spring Python
from springpython.context import DisposableObject
from springpython.jms.core import reserved_attributes, TextMessage
from springpython.util import TRACE1, synchronized
from springpython.jms import JMSException, WebSphereMQJMSException, \
    NoMessageAvailableException, DELIVERY_MODE_NON_PERSISTENT, \
    DELIVERY_MODE_PERSISTENT


# Don't pollute the caller's namespace
__all__ = ["WebSphereMQConnectionFactory"]


# Internal constants, don't touch.

# Some WMQ constants are not exposed by pymqi.
_WMQ_MQRFH_VERSION_2 = "\x00\x00\x00\x02"
_WMQ_DEFAULT_ENCODING = 273
_WMQ_DEFAULT_ENCODING_WIRE_FORMAT = pack("!l", _WMQ_DEFAULT_ENCODING)

# 1208 = UTF-8
_WMQ_DEFAULT_CCSID = 1208
_WMQ_DEFAULT_CCSID_WIRE_FORMAT = pack("!l", _WMQ_DEFAULT_CCSID)

# From cmqc.h
_WMQ_MQFMT_RF_HEADER_2 = "MQHRF2  "

# MQRFH_NO_FLAGS_WIRE is in cmqc.h
_WMQ_MQRFH_NO_FLAGS_WIRE_FORMAT = "\x00\x00\x00\x00"

# Java documentation says "214748364.7 seconds".
_WMQ_MAX_EXPIRY_TIME = 214748364.7

_WMQ_ID_PREFIX = "ID:"

# In current implementation, an mcd JMS folder is constant for every message
# sent, so let's build it here.

_mcd = etree.Element("mcd")
_msd = etree.Element("Msd")
_mcd.append(_msd)

# For now, it's always a TextMessage
_msd.text = "jms_text"

_msgbody = etree.Element("msgbody")
_msgbody.set("xmlns:xsi", "dummy") # We're using a dummy namespace
_msgbody.set("xsi:nil", "true")
_mcd.append(_msgbody)

# Clean up namespace.
del(_msd, _msgbody)


def unhexlify_wmq_id(wmq_id):
    """ Converts the WebSphere MQ generated identifier back to bytes,
    i.e. "ID:414d5120535052494e47505954484f4ecc90674a041f0020" -> "AMQ SPRINGPYTHON\xcc\x90gJ\x04\x1f\x00 ".
    """
    return unhexlify(wmq_id.replace(_WMQ_ID_PREFIX, "", 1))


class WebSphereMQConnectionFactory(DisposableObject):

    def __init__(self, queue_manager=None, channel=None, host=None, listener_port=None,
            cache_open_send_queues=True, cache_open_receive_queues=True,
            use_shared_connections=True, dynamic_queue_template="SYSTEM.DEFAULT.MODEL.QUEUE",
            ssl=False, ssl_cipher_spec=None, ssl_key_repository=None):
        self.queue_manager = queue_manager
        self.channel = channel
        self.host = host
        self.listener_port = listener_port

        self.use_shared_connections = use_shared_connections
        self.dynamic_queue_template = dynamic_queue_template

        # SSL support
        self.ssl = ssl
        self.ssl_cipher_spec = ssl_cipher_spec
        self.ssl_key_repository = ssl_key_repository

        self.logger = logging.getLogger("springpython.jms.factory.WebSphereMQConnectionFactory")

        import CMQC
        import pymqi

        self.CMQC = CMQC
        self.mq = pymqi

        self._open_send_queues_cache = {}
        self._open_receive_queues_cache = {}
        self._open_dynamic_queues_cache = {}

        self.cache_open_send_queues = cache_open_send_queues
        self.cache_open_receive_queues = cache_open_receive_queues

        self._is_connected = False
        self._disconnecting = False

        self.logger.log(TRACE1, "Finished __init__")

    @synchronized()
    def destroy(self):
        if self._is_connected:
            self._disconnecting = True
            try:
                self.logger.info("Deleting queues from caches")
                self._open_send_queues_cache.clear()
                self._open_receive_queues_cache.clear()
                self._open_dynamic_queues_cache.clear()
                self.logger.info("Caches cleared")
            except Exception, e:
                try:
                    self.logger.error("Could not clear the caches. Exception [%s]" % format_exc())
                except:
                    pass
            try:
                self.logger.info("Disconnecting from queue manager [%s]" % self.queue_manager)
                self.mgr.disconnect()
                self.logger.info("Disconnected from queue manager [%s]" % self.queue_manager)
            except Exception, e:
                try:
                    self.logger.error("Could not disconnect from queue manager [%s], exception [%s] " % (self.queue_manager,
                        format_exc()))
                except Exception:
                    pass

            self._is_connected = False

        else:
            self.logger.debug("Not connected, skipping cleaning up the resources")

    def get_connection_info(self):
        return "queue manager=[%s], channel=[%s], conn_name=[%s(%s)]" % (
            self.queue_manager, self.channel, self.host, self.listener_port)

    @synchronized()
    def _connect(self):
        if self._is_connected:
            return

        conn_name = "%s(%s)" % (self.host, self.listener_port)

        self.logger.info("Connecting to queue manager [%s], channel [%s]" \
            ", connection info [%s]" % (self.queue_manager, self.channel, conn_name))
        self.mgr = self.mq.QueueManager(None)

        sco = self.mq.sco()
        cd = self.mq.cd()
        cd.ChannelName = self.channel
        cd.ConnectionName = conn_name
        cd.ChannelType = self.CMQC.MQCHT_CLNTCONN
        cd.TransportType = self.CMQC.MQXPT_TCP

        if self.ssl:
            if not(self.ssl_cipher_spec and self.ssl_key_repository):
                msg = "SSL support requires setting both ssl_cipher_spec and ssl_key_repository"
                self.logger.error(msg)
                raise JMSException(msg)

            sco.KeyRepository = self.ssl_key_repository
            cd.SSLCipherSpec = self.ssl_cipher_spec

        if self.use_shared_connections:
            connect_options = self.CMQC.MQCNO_HANDLE_SHARE_BLOCK
        else:
            connect_options = self.CMQC.MQCNO_HANDLE_SHARE_NONE

        try:
            self.mgr.connectWithOptions(self.queue_manager, cd=cd, opts=connect_options, sco=sco)
        except self.mq.MQMIError, e:
            exc = WebSphereMQJMSException(e, e.comp, e.reason)
            raise exc
        except Exception, e:
            self.logger.error("Could not connect to queue manager, e=[%s]" % e)
            exc = WebSphereMQJMSException(e, None, None)
            raise exc
        else:
            self._is_connected = True
            self.logger.info("Successfully connected to queue manager [%s]" \
                ", channel [%s], connection info [%s]" % (self.queue_manager, self.channel, conn_name))

    def _get_queue_from_cache(self, destination, cache):
        lock = RLock()
        lock.acquire()
        try:
            # Will usually choose this path and find the queue here.
            if destination in cache:
                return cache[destination]
            else:
                self.logger.debug("Adding queue [%s] to the cache" % destination)
                cache[destination] = self.mq.Queue(self.mgr, destination, self.CMQC.MQOO_INPUT_SHARED | self.CMQC.MQOO_OUTPUT)
                self.logger.debug("Queue [%s] added to the cache" % destination)
                self.logger.log(TRACE1, "Cache contents [%s]" % cache)
                return cache[destination]
        finally:
            lock.release()

    def get_queue_for_sending(self, destination):
        if self.cache_open_send_queues:
            queue = self._get_queue_from_cache(destination, self._open_send_queues_cache)
        else:
            queue = self.mq.Queue(self.mgr, destination)

        return queue

    def get_queue_for_receiving(self, destination):
        if self.cache_open_receive_queues:
            queue = self._get_queue_from_cache(destination, self._open_receive_queues_cache)
        else:
            queue = self.mq.Queue(self.mgr, destination)

        return queue


    def send(self, message, destination):
        if self._disconnecting:
            self.logger.info("Connection factory disconnecting, aborting receive")
            return
        else:
            self.logger.log(TRACE1, "send -> not disconnecting")

        if not self._is_connected:
            self.logger.log(TRACE1, "send -> _is_connected1 %s" % self._is_connected)
            self._connect()
            self.logger.log(TRACE1, "send -> _is_connected2 %s" % self._is_connected)

        destination = self._strip_prefixes_from_destination(destination)

        # Will consist of an MQRFH2 header and the actual business payload.
        buff = StringIO()

        # Build the message descriptor (MQMD)
        md = self._build_md(message)

        # Create MQRFH2 header
        now = long(time() * 1000)
        mqrfh2jms = MQRFH2JMS().build_header(message, destination, self.CMQC, now)

        buff.write(mqrfh2jms)
        if message.text != None:
            buff.write(message.text.encode("utf-8"))

        body = buff.getvalue()
        buff.close()

        queue = self.get_queue_for_sending(destination)

        try:
            queue.put(body, md)
        except self.mq.MQMIError, e:
            self.logger.error("MQMIError in queue.put, e.comp [%s], e.reason [%s] " % (
                e.comp, e.reason))
            exc = WebSphereMQJMSException(e, e.comp, e.reason)
            raise exc

        if not self.cache_open_send_queues:
            queue.close()

        # Map the JMS headers overwritten by calling queue.put
        message.jms_message_id = _WMQ_ID_PREFIX + hexlify(md.MsgId)
        message.jms_priority = md.Priority
        message.jms_correlation_id = _WMQ_ID_PREFIX + hexlify(md.CorrelId)
        message.JMSXUserID = md.UserIdentifier
        message.JMSXAppID = md.PutApplName

        if md.PutDate and md.PutTime:
            message.jms_timestamp = self._get_jms_timestamp_from_md(md.PutDate.strip(), md.PutTime.strip())
            message.JMS_IBM_PutDate = md.PutDate.strip()
            message.JMS_IBM_PutTime = md.PutTime.strip()
        else:
            self.logger.warning("No md.PutDate and md.PutTime found, md [%r]" % repr(md))

        # queue.put has succeeded, so overwrite expiration time as well
        if message.jms_expiration != None:
            message.jms_expiration += now

        self.logger.debug("Successfully sent a message [%s], connection info [%s]" % (
            message, self.get_connection_info()))

        self.logger.log(TRACE1, "message [%s], body [%r], md [%r]" % (message, body, repr(md)))

    def receive(self, destination, wait_interval):
        if self._disconnecting:
            self.logger.info("Connection factory disconnecting, aborting receive")
            return
        else:
            self.logger.log(TRACE1, "receive -> not disconnecting")


        if not self._is_connected:
            self.logger.log(TRACE1, "receive -> _is_connected1 %s" % self._is_connected)
            self._connect()
            self.logger.log(TRACE1, "receive -> _is_connected2 %s" % self._is_connected)

        queue = self.get_queue_for_receiving(destination)

        try:
            # Default message descriptor ..
            md = self.mq.md()

            # .. and custom get message options
            gmo = self.mq.gmo()
            gmo.Options = self.CMQC.MQGMO_WAIT | self.CMQC.MQGMO_FAIL_IF_QUIESCING
            gmo.WaitInterval = wait_interval

            message = queue.get(None, md, gmo)

            return self._build_text_message(md, message)

        except self.mq.MQMIError, e:
            if e.reason == self.CMQC.MQRC_NO_MSG_AVAILABLE:
                text = "No message available for destination [%s], " \
                    "wait_interval [%s] ms" % (destination, wait_interval)
                raise NoMessageAvailableException(text)
            else:
                self.logger.log(TRACE1, "Exception caught in get, e.comp=[%s], e.reason=[%s]" % (e.comp, e.reason))
                exc = WebSphereMQJMSException(e, e.comp, e.reason)
                raise exc


    def open_dynamic_queue(self):
        if self._disconnecting:
            self.logger.info("Connection factory disconnecting, aborting open_dynamic_queue")
            return
        else:
            self.logger.log(TRACE1, "open_dynamic_queue -> not disconnecting")

        if not self._is_connected:
            self.logger.log(TRACE1, "open_dynamic_queue -> _is_connected1 %s" % self._is_connected)
            self._connect()
            self.logger.log(TRACE1, "open_dynamic_queue -> _is_connected2 %s" % self._is_connected)

        dynamic_queue = self.mq.Queue(self.mgr, self.dynamic_queue_template,
            self.CMQC.MQOO_INPUT_SHARED)

        # A bit hackish, but there's no other way to get its name.
        dynamic_queue_name = dynamic_queue._Queue__qDesc.ObjectName.strip()

        lock = RLock()
        lock.acquire()
        try:
            self._open_dynamic_queues_cache[dynamic_queue_name] = dynamic_queue
        finally:
            lock.release()

        self.logger.log(TRACE1, "Successfully created a dynamic queue, descriptor [%s]" % (
            dynamic_queue._Queue__qDesc))

        return dynamic_queue_name

    def close_dynamic_queue(self, dynamic_queue_name):
        if self._disconnecting:
            self.logger.info("Connection factory disconnecting, aborting close_dynamic_queue")
            return
        else:
            self.logger.log(TRACE1, "close_dynamic_queue -> not disconnecting")

        if not self._is_connected:
            # If we're not connected then all dynamic queues had been already closed.
            self.logger.log(TRACE1, "close_dynamic_queue -> _is_connected1 %s" % self._is_connected)
            return
        else:
            self.logger.log(TRACE1, "close_dynamic_queue -> _is_connected2 %s" % self._is_connected)
            lock = RLock()
            lock.acquire()
            try:
                dynamic_queue = self._open_dynamic_queues_cache[dynamic_queue_name]
                dynamic_queue.close()

                self._open_dynamic_queues_cache.pop(dynamic_queue_name, None)
                self._open_send_queues_cache.pop(dynamic_queue_name, None)
                self._open_receive_queues_cache.pop(dynamic_queue_name, None)

                self.logger.log(TRACE1, "Successfully closed a dynamic queue [%s]" % (
                    dynamic_queue_name))

            finally:
                lock.release()

    def _get_jms_timestamp_from_md(self, put_date, put_time):
        pattern = "%Y%m%d%H%M%S"
        centi = int(put_time[6:]) / 100.0

        strp = strptime(put_date + put_time[:6], pattern)
        mk = mktime(strp)

        return long((mk - altzone + centi) * 1000.0)


    def _build_text_message(self, md, message):
        self.logger.log(TRACE1, "Building a text message [%r], md [%r]" % (repr(message), repr(md)))

        mqrfh2 = MQRFH2JMS()
        mqrfh2.build_folders_and_payload_from_message(message)

        jms_folder = mqrfh2.folders.get("jms", None)
        mcd_folder = mqrfh2.folders.get("mcd", None)
        usr_folder = mqrfh2.folders.get("usr", None)

        # Create a message instance ..
        text_message = TextMessage()

        if usr_folder:
            for attr_name, attr_value in usr_folder.items():
                setattr(text_message, attr_name, str(attr_value))

        # .. set its JMS properties ..

        if jms_folder:
            if jms_folder.find("Dst") is not None:
                text_message.jms_destination = jms_folder.find("Dst").text.strip()

            if jms_folder.find("Exp") is not None:
                text_message.jms_expiration = long(jms_folder.find("Exp").text)
            else:
                text_message.jms_expiration = 0 # Same as in Java

            if jms_folder.find("Cid") is not None:
                text_message.jms_correlation_id = jms_folder.find("Cid").text

        if md.Persistence == self.CMQC.MQPER_NOT_PERSISTENT:
            text_message.jms_delivery_mode = DELIVERY_MODE_NON_PERSISTENT
        elif md.Persistence in(self.CMQC.MQPER_PERSISTENT, self.CMQC.MQPER_PERSISTENCE_AS_Q_DEF):
            text_message.jms_delivery_mode = DELIVERY_MODE_PERSISTENT
        else:
            text = "Don't know how to handle md.Persistence mode [%s]" % (md.Persistence)
            self.logger.error(text)
            exc = WebSphereMQJMSException(text)
            raise exc

        if md.ReplyToQ.strip():
            self.logger.log(TRACE1, "Found md.ReplyToQ=[%r]" % md.ReplyToQ)
            text_message.jms_reply_to = "queue://" + md.ReplyToQMgr.strip() + "/" + md.ReplyToQ.strip()

        text_message.jms_priority = md.Priority
        text_message.jms_message_id = _WMQ_ID_PREFIX + hexlify(md.MsgId)
        text_message.jms_timestamp = self._get_jms_timestamp_from_md(md.PutDate.strip(), md.PutTime.strip())
        text_message.jms_redelivered = bool(int(md.BackoutCount))

        text_message.JMSXUserID = md.UserIdentifier.strip()
        text_message.JMSXAppID = md.PutApplName.strip()
        text_message.JMSXDeliveryCount = md.BackoutCount
        text_message.JMSXGroupID = md.GroupId.strip()
        text_message.JMSXGroupSeq = md.MsgSeqNumber

        md_report_to_jms = {
            self.CMQC.MQRO_EXCEPTION: "Exception",
            self.CMQC.MQRO_EXPIRATION: "Expiration",
            self.CMQC.MQRO_COA: "COA",
            self.CMQC.MQRO_COD: "COD",
            self.CMQC.MQRO_PAN: "PAN",
            self.CMQC.MQRO_NAN: "NAN",
            self.CMQC.MQRO_PASS_MSG_ID: "Pass_Msg_ID",
            self.CMQC.MQRO_PASS_CORREL_ID: "Pass_Correl_ID",
            self.CMQC.MQRO_DISCARD_MSG: "Discard_Msg",
        }

        for report_name, jms_header_name in md_report_to_jms.iteritems():
            report_value = md.Report & report_name
            if report_value:
                header_value = report_value
            else:
                header_value = None

            setattr(text_message, "JMS_IBM_Report_" + jms_header_name, header_value)

        text_message.JMS_IBM_MsgType = md.MsgType
        text_message.JMS_IBM_Feedback = md.Feedback
        text_message.JMS_IBM_Format = md.Format.strip()
        text_message.JMS_IBM_PutApplType = md.PutApplType
        text_message.JMS_IBM_PutDate = md.PutDate.strip()
        text_message.JMS_IBM_PutTime = md.PutTime.strip()

        if md.MsgFlags & self.CMQC.MQMF_LAST_MSG_IN_GROUP:
            text_message.JMS_IBM_Last_Msg_In_Group = self.CMQC.MQMF_LAST_MSG_IN_GROUP
        else:
            text_message.JMS_IBM_Last_Msg_In_Group = None

        # .. and its payload too.
        if mqrfh2.payload:
            text_message.text = mqrfh2.payload

        return text_message

    def _strip_prefixes_from_destination(self, destination):
        if destination.startswith("queue:///"):
            return destination.replace("queue:///", "", 1)
        elif destination.startswith("queue://"):
            no_qm_dest = destination.replace("queue://", "", 1)
            no_qm_dest = no_qm_dest.split("/")[1:]
            return "/".join(no_qm_dest)
        else:
            return destination

    def _build_md(self, message):
        md = self.mq.md()

        md.Format = _WMQ_MQFMT_RF_HEADER_2
        md.CodedCharSetId = _WMQ_DEFAULT_CCSID
        md.Encoding = _WMQ_DEFAULT_ENCODING

        # Map JMS headers to MQMD

        if message.jms_correlation_id:
            if message.jms_correlation_id.startswith(_WMQ_ID_PREFIX):
                md.CorrelId = unhexlify_wmq_id(message.jms_correlation_id)
            else:
                md.CorrelId = message.jms_correlation_id.ljust(24)[:24]

        if message.jms_delivery_mode:

            if message.jms_delivery_mode == DELIVERY_MODE_NON_PERSISTENT:
                persistence = self.CMQC.MQPER_NOT_PERSISTENT
            elif message.jms_delivery_mode == DELIVERY_MODE_PERSISTENT:
                persistence = self.CMQC.MQPER_PERSISTENT
            else:
                info = "jms_delivery_mode should be equal to DELIVERY_MODE_NON_PERSISTENT or DELIVERY_MODE_PERSISTENT, not [%s]" % message.jms_delivery_mode
                self.logger.error(info)
                exc = JMSException(info)
                raise exc

            md.Persistence = persistence

        if message.jms_priority:
            md.Priority = message.jms_priority

        if message.jms_reply_to:
            md.ReplyToQ = message.jms_reply_to

            self.logger.log(TRACE1, ("Set jms_reply_to. md.ReplyToQ=[%r],"
                " message.jms_reply_to=[%r]" % (md.ReplyToQ, message.jms_reply_to)))

        # jms_expiration is in milliseconds, md.Expiry is in centiseconds.
        if message.jms_expiration:
            if message.jms_expiration / 1000 > _WMQ_MAX_EXPIRY_TIME:
                md.Expiry = self.CMQC.MQEI_UNLIMITED
            else:
                md.Expiry = message.jms_expiration / 10

        # WebSphere MQ provider-specific JMS headers

        jmsxgroupseq = getattr(message, "JMSXGroupSeq", None)
        if jmsxgroupseq != None:
            md.MsgSeqNumber = jmsxgroupseq
            md.MsgFlags |= self.CMQC.MQMF_MSG_IN_GROUP

        jmsxgroupid = getattr(message, "JMSXGroupID", None)
        if jmsxgroupid != None:
            if jmsxgroupid.startswith(_WMQ_ID_PREFIX):
                md.GroupId = unhexlify_wmq_id(jmsxgroupid)
            else:
                md.GroupId = jmsxgroupid.ljust(24)[:24]
            md.MsgFlags |= self.CMQC.MQMF_MSG_IN_GROUP

        for report_name in("Exception", "Expiration", "COA", "COD", "PAN",
            "NAN", "Pass_Msg_ID", "Pass_Correl_ID", "Discard_Msg"):

            report = getattr(message, "JMS_IBM_Report_" + report_name, None)
            if report != None:
                md.Report |= report

        # Doesn't make much sense to map feedback options as we're stuffed into
        # request messages (MQMT_REQUEST) not report messages (MQMT_REPORT)
        # but different types of messages are still possible to implement in
        # the future so let's leave it.

        jms_ibm_feedback = getattr(message, "JMS_IBM_Feedback", None)
        if jms_ibm_feedback != None:
            md.Feedback = jms_ibm_feedback

        jms_ibm_last_msg_in_group = getattr(message, "JMS_IBM_Last_Msg_In_Group", None)
        if jms_ibm_last_msg_in_group != None:
            md.MsgFlags |= self.CMQC.MQMF_LAST_MSG_IN_GROUP

        return md

class MQRFH2JMS(object):
    """ A class for representing a subset of MQRFH2, suitable for passing
    WebSphere MQ JMS headers around.
    """

    # 4 bytes - MQRFH_STRUC_ID
    # 4 bytes - _WMQ_MQRFH_VERSION_2
    # 4 bytes - the whole MQRFH2 header length
    # 4 bytes - Encoding
    # 4 bytes - CodedCharacterSetId
    # 8 bytes - MQFMT_STRING
    # 4 bytes - MQRFH_NO_FLAGS
    # 4 bytes - NameValueCCSID
    FIXED_PART_LENGTH = 36

    # MQRFH2 folder length must be a multiple of 4.
    FOLDER_LENGTH_MULTIPLE = 4

    # Size of a folder header is always 4 bytes.
    FOLDER_SIZE_HEADER_LENGTH = 4

    def __init__(self):
        self.folders = {}
        self.payload = None

        self.logger = logging.getLogger("springpython.jms.factory.MQRFH2JMS")

    def _pad_folder(self, folder):
        """ Pads the folder to a multiple of 4, as required by WebSphere MQ.
        """
        folder_len = len(folder)

        if folder_len % MQRFH2JMS.FOLDER_LENGTH_MULTIPLE == 0:
            return folder
        else:
            padding = MQRFH2JMS.FOLDER_LENGTH_MULTIPLE - folder_len % MQRFH2JMS.FOLDER_LENGTH_MULTIPLE
            return folder.ljust(folder_len + padding)

    def build_folders_and_payload_from_message(self, message):
        total_mqrfh2_length = unpack("!l", message[8:12])[0]

        mqrfh2 = message[MQRFH2JMS.FIXED_PART_LENGTH:total_mqrfh2_length]
        self.payload = message[MQRFH2JMS.FIXED_PART_LENGTH + len(mqrfh2):]

        self.logger.log(TRACE1, "message [%r]" % message)
        self.logger.log(TRACE1, "mqrfh2 [%r]" % mqrfh2)
        self.logger.log(TRACE1, "self.payload [%r]" % self.payload)

        left = mqrfh2
        while left:
            current_folder_length = unpack("!l", left[:4])[0]
            raw_folder = left[MQRFH2JMS.FOLDER_SIZE_HEADER_LENGTH:MQRFH2JMS.FOLDER_SIZE_HEADER_LENGTH + current_folder_length]

            self.logger.log(TRACE1, "raw_folder [%r]" % raw_folder)
            self.build_folder(raw_folder)

            left = left[MQRFH2JMS.FOLDER_SIZE_HEADER_LENGTH  + current_folder_length:]

    def build_folder(self, raw_folder):

        # Java JMS sends folders with unbound prefixes, i.e. <msgbody xsi:nil="true"></msgbody>
        # which is in no way a valid XML so we have to insert the prefix ourselves
        # in order to avoid parser bailing out with an ExpatError. I can't think
        # of any other way to work around it if we'd like to treat folders as
        # XML(-like) structures.

        if 'xsi:nil="true"' in raw_folder and not 'xmlns' in raw_folder:
            self.logger.log(TRACE1, "Binding xsi:nil to a dummy namespace [%s]" % raw_folder)
            raw_folder = raw_folder.replace('xsi:nil="true"', 'xmlns:xsi="dummy" xsi:nil="true"')
            self.logger.log(TRACE1, "raw_folder after binding [%s]" % raw_folder)

        folder = etree.fromstring(raw_folder)
        root_name = folder.tag

        if root_name in("mcd", "jms", "usr"):
            self.folders[root_name] = folder
        else:
            self.logger.warn("Ignoring unrecognized JMS folder [%s]=[%s]" % (root_name, raw_folder))


    def build_header(self, message, queue_name, CMQC, now):
        self.folders["mcd"] = _mcd
        self.add_jms(message, queue_name, now)
        self.add_usr(message)

        mcd = self._pad_folder(etree.tostring(self.folders["mcd"]))
        jms = self._pad_folder(etree.tostring(self.folders["jms"]))

        if "usr" in self.folders:
            usr = self._pad_folder(etree.tostring(self.folders["usr"]))
            usr_len = len(usr)
        else:
            usr_len = 0

        mcd_len = len(mcd)
        jms_len = len(jms)

        total_header_length = 0
        total_header_length += MQRFH2JMS.FIXED_PART_LENGTH

        # Each folder has a 4-byte header describing its length,
        # hence the "len(self.folders) * 4" below.
        variable_part_length = len(self.folders) * 4 + mcd_len + jms_len + usr_len

        total_header_length += variable_part_length

        buff = StringIO()

        buff.write(CMQC.MQRFH_STRUC_ID)
        buff.write(_WMQ_MQRFH_VERSION_2)
        buff.write(pack("!l", total_header_length))
        buff.write(_WMQ_DEFAULT_ENCODING_WIRE_FORMAT)
        buff.write(_WMQ_DEFAULT_CCSID_WIRE_FORMAT)
        buff.write(CMQC.MQFMT_STRING)
        buff.write(_WMQ_MQRFH_NO_FLAGS_WIRE_FORMAT)
        buff.write(_WMQ_DEFAULT_CCSID_WIRE_FORMAT)
        buff.write(pack("!l", mcd_len))
        buff.write(mcd)
        buff.write(pack("!l", jms_len))
        buff.write(jms)

        if "usr" in self.folders:
            buff.write(pack("!l", usr_len))
            buff.write(usr)

        value = buff.getvalue()
        buff.close()

        return value

    def add_jms(self, message, queue_name, now):

        jms = etree.Element("jms")
        dst = etree.Element("Dst")
        tms = etree.Element("Tms")
        dlv = etree.Element("Dlv")

        jms.append(dst)
        jms.append(tms)
        jms.append(dlv)

        tms.text = unicode(now)
        dst.text = u"queue:///" + queue_name
        dlv.text = unicode(message.jms_delivery_mode)

        if message.jms_expiration:
            exp = etree.Element("Exp")
            exp.text = unicode(now + message.jms_expiration)
            self.logger.log(TRACE1, "jms.Exp [%r]" % exp.text)
            jms.append(exp)

        if message.jms_priority:
            pri = etree.Element("Pri")
            pri.text = unicode(message.jms_priority)
            self.logger.log(TRACE1, "jms.Pri [%r]" % pri.text)
            jms.append(pri)

        if message.jms_correlation_id:
            cid = etree.Element("Cid")
            cid.text = unicode(message.jms_correlation_id)
            self.logger.log(TRACE1, "jms.Cid [%r]" % cid.text)
            jms.append(cid)

        self.folders["jms"] = jms

    def add_usr(self, message):
        user_attrs = set(dir(message)) - reserved_attributes
        self.logger.log(TRACE1, "user_attrs [%s]" % user_attrs)

        if user_attrs:
            usr = etree.Element("usr")

            for user_attr in user_attrs:

                user_attr_value = getattr(message, user_attr)

                # Some values are integers, e.g. delivery_mode
                if isinstance(user_attr_value, basestring):
                    user_attr_value = escape(user_attr_value)

                # Create a JMS attribute and set its value.
                user_attr = etree.Element(unicode(user_attr))
                user_attr.text = unicode(user_attr_value)
                usr.append(user_attr)

            self.folders["usr"] = usr



# -*- coding: utf-8 -*-

# stdlib
import httplib
import socket
import ssl

class CAValidatingHTTPSConnection(httplib.HTTPConnection):
    """ This class allows communication via SSL/TLS and takes Certificate Authorities
    into account.
    """

    def __init__(self, host, port=None, ca_certs=None, keyfile=None, certfile=None,
                 cert_reqs=None, strict=None, ssl_version=None,
                 timeout=None):
        httplib.HTTPConnection.__init__(self, host, port, strict, timeout)

        self.ca_certs = ca_certs
        self.keyfile = keyfile
        self.certfile = certfile
        self.cert_reqs = cert_reqs
        self.ssl_version = ssl_version

    def connect(self):
        """ Connect to a host on a given (SSL/TLS) port.
        """
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        self.sock = self.wrap_socket(sock)

    def wrap_socket(self, sock):
        """ Gets a socket object and wraps it into an SSL/TLS-aware one. May be
        overridden in subclasses if the wrapping process needs to be customized.
        """
        return ssl.wrap_socket(sock, self.keyfile, self.certfile,
                                    ca_certs=self.ca_certs, cert_reqs=self.cert_reqs,
                                    ssl_version=self.ssl_version)

class CAValidatingHTTPS(httplib.HTTP):
    """ A subclass of httplib.HTTP which is aware of Certificate Authorities
    used in SSL/TLS transactions.
    """
    _connection_class = CAValidatingHTTPSConnection

    def __init__(self, host=None, port=None, strict=None, ca_certs=None, keyfile=None, certfile=None,
                 cert_reqs=None, ssl_version=None, timeout=None):
        self._setup(self._connection_class(host, port, ca_certs, keyfile, certfile,
                                           cert_reqs, strict, ssl_version, timeout))


"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""

import logging
from springpython.context import scope

class ObjectContainer(object):
    """
    ObjectContainer is a container which uses multiple Config objects to read sources of
    object definitions. When a object is requested from this container, it may optionally
    pull the object from a scoped cache. If there is no stored copy of the object, it
    uses the scanned definition and its associated ObjectFactory to create an instance. It
    can then optionally store it in a scoped cache for future usage (e.g. singleton).
    
    Object definitions are stored in the container in a neutral format, decoupling the
    container entirely from the original source location. This means that XML, python code,
    and other formats may all contain definitions. By the time they
    reach this container, it doesn't matter what their original format was when a object
    instance is needed. NOTE: This explicitly means that one object in one source
    can refer to another object in another source OF ANY FORMAT as a property.
    """
    def __init__(self, config = None):
        self.logger = logging.getLogger("springpython.container.ObjectContainer")

        if config is None:
            self.configs = []
        elif isinstance(config, list):
            self.configs = config
        else:
            self.configs = [config]

        self.object_defs = {}
    
        for configuration in self.configs:
            self.logger.debug("=== Scanning configuration %s for object definitions ===" % configuration)
            for object_def in configuration.read_object_defs():
                if object_def.id not in self.object_defs:
                    self.logger.debug("%s object definition does not exist. Adding to list of definitions." % object_def.id)
                else:
                    self.logger.debug("Overriding previous definition of %s" % object_def.id)
                self.object_defs[object_def.id] = object_def

        self.logger.debug("=== Done reading object definitions. ===")

        self.objects = {}

    def get_object(self, name, ignore_abstract=False):
        """
        This function attempts to find the object in the singleton cache. If not found, 
        delegates to _create_object in order to hunt for the definition, and request a
        object factory to generate one.
        """
        try:
            object_def = self.object_defs[name]
            if object_def.abstract and not ignore_abstract:
                raise AbstractObjectException("Object [%s] is an abstract one." % name)
                
            return self.objects[name]
            
        except KeyError, e:
            self.logger.debug("Did NOT find object '%s' in the singleton storage." % name)
            try:
                object_def = self.object_defs[name]
                if object_def.abstract and not ignore_abstract:
                    raise AbstractObjectException("Object [%s] is an abstract one." % name)
                
                comp = self._create_object(object_def)
                
                # Evaluate any scopes, and store appropriately.
                if self.object_defs[name].scope == scope.SINGLETON:
                    self.objects[name] = comp
                    self.logger.debug("Stored object '%s' in container's singleton storage" % name)
                elif self.object_defs[name].scope == scope.PROTOTYPE:
                    pass
                else:
                    raise InvalidObjectScope("Don't know how to handle scope %s" % self.object_defs[name].scope)
                
                return comp
            except KeyError, e:
                self.logger.error("Object '%s' has no definition!" % name)
                raise e
            
    def _get_constructors_pos(self, object_def):
        """
        This function iterates over the positional constructors, and assembles their values into a list.
        In this situation, the order as read from the XML should be the order expected by the class
        definition.
        """
        return tuple([constr.get_value(self) for constr in object_def.pos_constr
                      if hasattr(constr, "get_value")])

    def _get_constructors_kw(self, kwargs):
        """
        This function iterates over the named constructors, and assembles their values into a list.
        In this situation, each argument is associated with a name, and due to unicode format provided
        by the XML parser, requires conversion into a new dictionary.
        """
        return dict([(key, kwargs[key].get_value(self)) for key in kwargs
                     if hasattr(kwargs[key], "get_value")])


    def _create_object(self, object_def):
        """
        If the object isn't stored in any scoped cache, and must instead be created, this method
        takes all the steps to read the object's definition, res it up, and store it in the appropriate
        scoped cache.
        """
        self.logger.debug("Creating an instance of %s" % object_def)
        
        [constr.prefetch(self) for constr in object_def.pos_constr if hasattr(constr, "prefetch")]
        [constr.prefetch(self) for constr in object_def.named_constr.values() if hasattr(constr, "prefetch")]
        [prop.prefetch(self) for prop in object_def.props if hasattr(prop, "prefetch")]
        
        # Res up an instance of the object, with ONLY constructor-based properties set.
        obj = object_def.factory.create_object(self._get_constructors_pos(object_def),
                                               self._get_constructors_kw(object_def.named_constr))

        # Fill in the other property values.
        [prop.set_value(obj, self) for prop in object_def.props if hasattr(prop, "set_value")]
        
        return obj
        
        
class AbstractObjectException(Exception):
    """ Raised when the user's code tries to get an abstract object from
    the container.
    """
    
class InvalidObjectScope(Exception):
    pass



"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
import logging
import Pyro.core
from springpython.context import InitializingObject
from springpython.remoting.pyro import PyroDaemonHolder

class PyroServiceExporter(InitializingObject):
    """
    This class will expose an object using Pyro. It requires that a daemon thread
    be up and running in order to receive requests and allow dispatching to the exposed
    object.
    """
    def __init__(self, service = None, service_name = None, service_host = "localhost", service_port = 7766):
        self.logger = logging.getLogger("springpython.remoting.pyro.PyroServiceExporter")
        self.service = service
        self.service_name = service_name
        self.service_host = service_host
        self.service_port = service_port
        self._pyro_thread = None
        
    def __del__(self):
        """
        When the service exporter goes out of scope and is garbage collected, the
        service must be deregistered.
        """
        PyroDaemonHolder.deregister(self.service_name, self.service_host, self.service_port)
        
    def __setattr__(self, name, value):
        """
        Only the explicitly listed attributes can be assigned values. Everything else is passed through to
        the actual service.
        """
        if name in ["logger", "service", "service_name", "service_host", "service_port", "_pyro_thread"]:
            self.__dict__[name] = value
        else:
            object.__setattr__(self, name, value)

    def after_properties_set(self):
        if self.service is None: raise Exception("service must NOT be None")
        if self.service_name is None: raise Exception("service_name must NOT be None")
        if self.service_host is None: raise Exception("service_host must NOT be None")
        if self.service_port is None: raise Exception("service_port must NOT be None")
        self.logger.debug("Exporting %s as a Pyro service at %s:%s" % (self.service_name, self.service_host, self.service_port))
        pyro_obj = Pyro.core.ObjBase()
        pyro_obj.delegateTo(self.service)
        PyroDaemonHolder.register(pyro_obj, self.service_name, self.service_host, self.service_port)
            
class PyroProxyFactory(object):
    """
    This is wrapper around a Pyro client proxy. The idea is to inject this object with a 
    Pyro service_url, which in turn generates a Pyro client proxy. After that, any
    method calls or attribute accessses will be forwarded to the Pyro client proxy.
    """
    def __init__(self):
        self.__dict__["client_proxy"] = None
        self.__dict__["service_url"] = None
        
    def __setattr__(self, name, value):
        if name == "service_url":
            self.__dict__["service_url"] = value
        else:
            setattr(self.client_proxy, name, value)
    
    def __getattr__(self, name):
        if name in ["service_url"]:
            return self.__dict__[name]
        elif name in ["post_process_before_initialization", "post_process_after_initialization"]:
            raise AttributeError, name
        else:
            if self.client_proxy is None:
                self.__dict__["client_proxy"] = Pyro.core.getProxyForURI(self.service_url)
            return getattr(self.client_proxy, name)

class Pyro4ServiceExporter(InitializingObject):
    """
    This class will expose an object using Pyro. It requires that a daemon thread
    be up and running in order to receive requests and allow dispatching to the exposed
    object.
    """
    def __init__(self, service = None, service_name = None, service_host = "localhost", service_port = 7766):
        self.logger = logging.getLogger("springpython.remoting.pyro.Pyro4ServiceExporter")
        self.service = service
        self.service_name = service_name
        self.service_host = service_host
        self.service_port = service_port
        self._pyro_thread = None
        
    def __del__(self):
        """
        When the service exporter goes out of scope and is garbage collected, the
        service must be deregistered.
        """
        from springpython.remoting.pyro import Pyro4DaemonHolder
        Pyro4DaemonHolder.deregister(self.service_name, self.service_host, self.service_port)
        
    def __setattr__(self, name, value):
        """
        Only the explicitly listed attributes can be assigned values. Everything else is passed through to
        the actual service.
        """
        if name in ["logger", "service", "service_name", "service_host", "service_port", "_pyro_thread"]:
            self.__dict__[name] = value
        else:
            object.__setattr__(self, name, value)

    def after_properties_set(self):
        import Pyro4
        from springpython.remoting.pyro import Pyro4DaemonHolder
        if self.service is None: raise Exception("service must NOT be None")
        if self.service_name is None: raise Exception("service_name must NOT be None")
        if self.service_host is None: raise Exception("service_host must NOT be None")
        if self.service_port is None: raise Exception("service_port must NOT be None")
        self.logger.debug("Exporting %s as a Pyro service at %s:%s" % (self.service_name, self.service_host, self.service_port))
	wrapping_obj = PyroWrapperObj(self.service)
        Pyro4DaemonHolder.register(wrapping_obj, self.service_name, self.service_host, self.service_port)
            
class PyroWrapperObj(object):
    def __init__(self, delegate):
        self.delegate = delegate

    def __getattr__(self, name):
        if name in ["__pyroInvoke", "__call__", "_pyroId", "_pyroDaemon", "delegate"]:
            return self.__dict__[name]
        else:
            return getattr(self.delegate, name)

    def __setattr__(self, name, value):
        if name in ["__pyroInvoke", "__call__", "_pyroId", "_pyroDaemon", "delegate"]:
            self.__dict__[name] = value
        else:
            setattr(self.delegate, name, value)

class Pyro4ProxyFactory(object):
    """
    This is wrapper around a Pyro client proxy. The idea is to inject this object with a 
    Pyro service_url, which in turn generates a Pyro client proxy. After that, any
    method calls or attribute accessses will be forwarded to the Pyro client proxy.
    """
    def __init__(self):
        self.__dict__["client_proxy"] = None
        self.__dict__["service_url"] = None
        
    def __setattr__(self, name, value):
        if name == "service_url":
            self.__dict__["service_url"] = value
        else:
            setattr(self.client_proxy, name, value)
    
    def __getattr__(self, name):
        import Pyro4
        if name in ["service_url"]:
            return self.__dict__[name]
        elif name in ["post_process_before_initialization", "post_process_after_initialization"]:
            raise AttributeError, name
        else:
            if self.client_proxy is None:
                self.__dict__["client_proxy"] = Pyro4.Proxy(self.service_url)
            return getattr(self.client_proxy, name)




"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
import logging
import re
import sys
import types

class ConnectionFactory(object):
    def __init__(self, acceptable_types):
        self.__db = None
        self.acceptable_types = acceptable_types

    """This interface defines an object that is able to make database connections.
    This allows database connections to be defined inside application contexts, and
    fed to DAO and DatabaseTemplates."""
    def connect(self):
        raise NotImplementedError()

    def getConnection(self):
        if self.__db is None:
            self.__db = self.connect()
        return self.__db

    def close(self):
        "Need to offer API call to close the connection to the database."
        if self.__db is not None:
            self.__db.close()
            self.__db = None

    def commit(self):
        if self.in_transaction():
            self.getConnection().commit()

    def rollback(self):
        if self.in_transaction():
            self.getConnection().rollback()

    def in_transaction(self):
        raise NotImplementedError()

    def count_type(self):
        raise NotImplementedError()
    
    def convert_sql_binding(self, sql_query):
        """This is to help Java users migrate to Python. Java notation defines binding variables
        points with '?', while Python uses '%s', and this method will convert from one format
        to the other."""
        return re.sub(pattern="\?", repl="%s", string=sql_query)

class MySQLConnectionFactory(ConnectionFactory):
    def __init__(self, username = None, password = None, hostname = None, db = None):
        ConnectionFactory.__init__(self, [types.TupleType])
        self.username = username
        self.password = password
        self.hostname = hostname
        self.db = db
        
    def connect(self):
        """The import statement is delayed so the library is loaded ONLY if this factory is really used."""
        import MySQLdb
        return MySQLdb.connect(self.hostname, self.username, self.password, self.db)

    def in_transaction(self):
        return True

    def count_type(self):
        return types.LongType

class PgdbConnectionFactory(ConnectionFactory):
    def __init__(self, user = None, password = None, host = None, database = None):
        ConnectionFactory.__init__(self, [types.TupleType])
        self.user = user
        self.password = password
        self.host = host
        self.database = database
        
    def connect(self):
        """The import statement is delayed so the library is loaded ONLY if this factory is really used."""
        import pgdb
        return pgdb.connect(user=self.user, password=self.password, database=self.database, host=self.host)

    def in_transaction(self):
        return True

    def count_type(self):
        return types.LongType

class Sqlite3ConnectionFactory(ConnectionFactory):
    def __init__(self, db = None, check_same_thread=True):
        ConnectionFactory.__init__(self, [types.TupleType])
        self.db = db
        self.check_same_thread = check_same_thread
        self.using_sqlite3 = True

    def connect(self):
        """The import statement is delayed so the library is loaded ONLY if this factory is really used."""
        try:
            import sqlite3
            return sqlite3.connect(self.db, check_same_thread=self.check_same_thread)               
        except:
            import sqlite
            self.using_sqlite3 = False
            return sqlite.connect(self.db, check_same_thread=self.check_same_thread)

    def in_transaction(self):
        return True

    def count_type(self):
        return types.IntType

    def convert_sql_binding(self, sql_query):
        if self.using_sqlite3:
            """sqlite3 uses the ? notation, like Java's JDBC."""
            return re.sub(pattern="%s", repl="?", string=sql_query)
        else:
            """Older versions of sqlite use the %s notation"""
            return re.sub(pattern="\?", repl="%s", string=sql_query)

class cxoraConnectionFactory(ConnectionFactory):
    def __init__(self, username = None, password = None, hostname = None, db = None):
        ConnectionFactory.__init__(self, [types.DictType])
        self.username = username
        self.password = password
        self.hostname = hostname
        self.db = db

    def connect(self):
        """The import statement is delayed so the library is loaded ONLY if this factory is really used."""
        import cx_Oracle
        return cx_Oracle.connect(self.username, self.password, self.db)
        
class SQLServerConnectionFactory(ConnectionFactory):
    def __init__(self, **odbc_info):
        ConnectionFactory.__init__(self, [types.TupleType])
        self.odbc_info = odbc_info

    def connect(self):
        """The import statement is delayed so the library is loaded ONLY if this factory is really used."""
        import pyodbc
        odbc_info = ";".join(["%s=%s" % (key, value) for key, value in self.odbc_info.items()])
        return pyodbc.connect(odbc_info)
        
    def in_transaction(self):
        return True
        
    def count_type(self):
        return types.IntType
        
    def convert_sql_binding(self, sql_query):
        """SQL Server expects parameters to be passed as question marks."""
        return re.sub(pattern="%s", repl="?", string=sql_query)



# -*- coding: utf-8 -*-

# stdlib
import httplib
import logging
import socket
import ssl
import sys
import traceback

from SimpleXMLRPCServer import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler
from xmlrpclib import ServerProxy, Error, Transport

# Spring Python
from springpython.remoting.http import CAValidatingHTTPS

__all__ = ["VerificationException", "SSLServer", "SSLClient"]

class VerificationException(Exception):
    """ Raised when the verification of a certificate's fields fails.
    """

# ##############################################################################
# Server
# ##############################################################################

class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ("/", "/RPC2",)

    def setup(self):
        self.connection = self.request # for doPOST
        self.rfile = socket._fileobject(self.request, "rb", self.rbufsize)
        self.wfile = socket._fileobject(self.request, "wb", self.wbufsize)

class SSLServer(object, SimpleXMLRPCServer):
    def __init__(self, host=None, port=None, keyfile=None, certfile=None,
                 ca_certs=None, cert_reqs=ssl.CERT_NONE, ssl_version=ssl.PROTOCOL_TLSv1,
                 do_handshake_on_connect=True, suppress_ragged_eofs=True, ciphers=None,
                 log_requests=True, **kwargs):

        SimpleXMLRPCServer.__init__(self, (host, port), requestHandler=RequestHandler)
        self.logger = logging.getLogger(self.__class__.__name__)

        self.keyfile = keyfile
        self.certfile = certfile
        self.ca_certs = ca_certs
        self.cert_reqs = cert_reqs
        self.ssl_version = ssl_version
        self.do_handshake_on_connect = do_handshake_on_connect
        self.suppress_ragged_eofs = suppress_ragged_eofs
        self.ciphers = ciphers

        # Looks awkward to use camelCase here but that's what SimpleXMLRPCRequestHandler
        # expects.
        self.logRequests = log_requests

        # 'verify_fields' is taken from kwargs to allow for adding more keywords
        # in future versions.
        self.verify_fields = kwargs.get("verify_fields")

        self.register_functions()

    def get_request(self):
        """ Overridden from SocketServer.TCPServer.get_request, wraps the socket in
        an SSL context.
        """
        sock, from_addr = self.socket.accept()

        # 'ciphers' argument is new in 2.7 and we must support 2.6 so add it
        # to kwargs conditionally, depending on the Python version.

        kwargs = {"keyfile":self.keyfile, "certfile":self.certfile,
                    "server_side":True, "cert_reqs":self.cert_reqs, "ssl_version":self.ssl_version,
                    "ca_certs":self.ca_certs, "do_handshake_on_connect":self.do_handshake_on_connect,
                    "suppress_ragged_eofs":self.suppress_ragged_eofs}

        if sys.version_info >= (2, 7):
            kwargs["ciphers"] = self.ciphers

        sock  = ssl.wrap_socket(sock, **kwargs)

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug("get_request cert='%s', from_addr='%s'" % (sock.getpeercert(), from_addr))

        return sock, from_addr

    def verify_request(self, sock, from_addr):
        """ Overridden from SocketServer.TCPServer.verify_request, adds validation of the
        other side's certificate fields.
        """
        try:
            if self.verify_fields:

                cert = sock.getpeercert()
                if not cert:
                    msg = "Couldn't verify fields, peer didn't send the certificate, from_addr='%s'" % (from_addr,)
                    raise VerificationException(msg)

                allow_peer, reason = self.verify_peer(cert, from_addr)
                if not allow_peer:
                    self.logger.error(reason)
                    sock.close()
                    return False

        except Exception, e:

            # It was either an error on our side or the client didn't send the
            # certificate even though self.cert_reqs was CERT_OPTIONAL (it couldn't
            # have been CERT_REQUIRED because we wouldn't have got so far, the
            # session would've been terminated much earlier in ssl.wrap_socket call).
            # Regardless of the reason we cannot accept the client in that case.

            msg = "Verification error='%s', cert='%s', from_addr='%s'" % (
                traceback.format_exc(e), sock.getpeercert(), from_addr)
            self.logger.error(msg)

            sock.close()
            return False

        return True

    def verify_peer(self, cert, from_addr):
        """ Verifies the other side's certificate. May be overridden in subclasses
        if the verification process needs to be customized.
        """

        subject = cert.get("subject")
        if not subject:
            msg = "Peer certificate doesn't have the 'subject' field, cert='%s'" % cert
            raise VerificationException(msg)

        subject = dict(elem[0] for elem in subject)

        for verify_field in self.verify_fields:

            expected_value = self.verify_fields[verify_field]
            cert_value = subject.get(verify_field, None)

            if not cert_value:
                reason = "Peer didn't send the '%s' field, subject fields received '%s'" % (
                    verify_field, subject)
                return False, reason

            if expected_value != cert_value:
                reason = "Expected the subject field '%s' to have value '%s' instead of '%s', subject='%s'" % (
                    verify_field, expected_value, cert_value, subject)
                return False, reason

        return True, None

    def register_functions(self):
        raise NotImplementedError("Must be overridden by subclasses")

# ##############################################################################
# Client
# ##############################################################################

class SSLClientTransport(Transport):
    """ Handles an HTTPS transaction to an XML-RPC server.
    """

    user_agent = "SSL XML-RPC Client (by http://springpython.webfactional.com)"

    def __init__(self, ca_certs=None, keyfile=None, certfile=None, cert_reqs=None,
                 ssl_version=None, timeout=None, strict=None):

        self.ca_certs = ca_certs
        self.keyfile = keyfile
        self.certfile = certfile
        self.cert_reqs = cert_reqs
        self.ssl_version = ssl_version
        self.timeout = timeout
        self.strict = strict

        Transport.__init__(self)

    def make_connection(self, host):
        return CAValidatingHTTPS(host, strict=self.strict, ca_certs=self.ca_certs,
                keyfile=self.keyfile, certfile=self.certfile, cert_reqs=self.cert_reqs,
                ssl_version=self.ssl_version, timeout=self.timeout)

class SSLClient(ServerProxy):
    def __init__(self, uri=None, ca_certs=None, keyfile=None, certfile=None,
                 cert_reqs=ssl.CERT_REQUIRED, ssl_version=ssl.PROTOCOL_TLSv1,
                 transport=None, encoding=None, verbose=0, allow_none=0, use_datetime=0,
                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT, strict=None):

        if not transport:
            _transport=SSLClientTransport(ca_certs, keyfile, certfile, cert_reqs,
                                         ssl_version, timeout, strict)
        else:
            _transport=transport(ca_certs, keyfile, certfile, cert_reqs,
                                         ssl_version, timeout, strict)


        ServerProxy.__init__(self, uri, _transport, encoding, verbose,
                        allow_none, use_datetime)

        self.logger = logging.getLogger(self.__class__.__name__)


"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import re
import types
import inspect
import logging

from _config_base import *
from springpython.context import scope
from decorator import decorator, partial
from springpython.context import ApplicationContextAware
from springpython.factory import PythonObjectFactory
from springpython.factory import ReflectiveObjectFactory
from springpython.container import InvalidObjectScope

yaml_mappings = {
    "str":"types.StringType", "unicode":"types.UnicodeType",
    "int":"types.IntType", "long":"types.LongType",
    "float":"types.FloatType", "decimal":"decimal.Decimal",
    "bool":"types.BooleanType", "complex":"types.ComplexType",
    "list":"types.ListType", "tuple":"types.TupleType",
    "dict":"types.DictType",
}

class YamlConfig(Config):
    """
    YamlConfig provides an alternative YAML-based version of objects.
    """
    def __init__(self, config_location):
        if isinstance(config_location, list):
            self.config_location = config_location
        else:
            self.config_location = [config_location]
        self.logger = logging.getLogger("springpython.config.YamlConfig")

        # By making this an instance-based property (instead of function local), inner object
        # definitions can add themselves to the list in the midst of parsing an input.
        self.objects = []

    def read_object_defs(self):
        import yaml

        self.logger.debug("==============================================================")
        # Reset, in case the file is re-read
        self.objects = []
        for config in self.config_location:
            self.logger.debug("* Parsing %s" % config)
            stream = file(config)
            doc = yaml.load(stream)
            self.logger.debug(doc)

            # A dictionary of abstract objects, keyed by their IDs, used in
            # traversing the hierarchies of parents; built upfront here for
            # convenience.
            abstract_objects = {}
            for object in doc["objects"]:
                if "abstract" in object:
                    abstract_objects[object["object"]] = object

            for object in doc["objects"]:
                if not "class" in object and not "parent" in object:
                    self._map_custom_class(object, yaml_mappings)

                elif "parent" in object:
                    # Children are added to self.objects during the children->abstract parents traversal.
                    pos_constr = self._get_pos_constr(object)
                    named_constr = self._get_named_constr(object)
                    props = self._get_props(object)
                    self._traverse_parents(object, object, pos_constr, named_constr, props, abstract_objects)
                    continue

                self._print_obj(object)
                self.objects.append(self._convert_object(object))

        self.logger.debug("==============================================================")
        self.logger.debug("objects = %s" % self.objects)
        return self.objects

    def _map_custom_class(self, obj, mappings):
        """ Enrich the object's attributes and make it look to the rest of
        YamlConfig as if the object had all of them right in the definition.
        """
        for class_name in mappings:
            if class_name in obj:
                self.logger.debug("Found a matching type: %s -> %s" % (obj["object"],
                    class_name))

                obj["class"] = mappings[class_name]
                obj["constructor-args"] = [obj[class_name]]
                break
        else:
            self.logger.warning("No matching type found for object %s" % obj)

    def _traverse_parents(self, leaf, child, pos_constr,
                            named_constr, props, abstract_objects):

        parent = abstract_objects[child["parent"]]

        # At this point we only build up the lists of parameters but we don't create
        # the object yet because the current parent object may still have its
        # own parent.

        # Positional constructors

        parent_pos_constrs = self._get_pos_constr(parent)

        # Make sure there are as many child positional parameters as there
        # are in the parent's list.

        len_pos_constr = len(pos_constr)
        len_parent_pos_constrs = len(parent_pos_constrs)

        if len_pos_constr < len_parent_pos_constrs:
            pos_constr.extend([None] * (len_parent_pos_constrs - len_pos_constr))

        for idx, parent_pos_constr in enumerate(parent_pos_constrs):
            if not pos_constr[idx]:
                pos_constr[idx] = parent_pos_constr

        # Named constructors
        child_named_constrs = named_constr
        parent_named_constrs = self._get_named_constr(parent)

        for parent_named_constr in parent_named_constrs:
            if parent_named_constr not in child_named_constrs:
                named_constr[parent_named_constr] = parent_named_constrs[parent_named_constr]

        # Properties
        child_props = [prop.name for prop in props]
        parent_props = self._get_props(parent)

        for parent_prop in parent_props:
            if parent_prop.name not in child_props:
                props.append(parent_prop)

        if "parent" in parent:
            self._traverse_parents(leaf, parent, pos_constr, named_constr, props, abstract_objects)
        else:
            # Now we know we can create an object out of all the accumulated values.

            # The object's class is its topmost parent's class.
            class_ = parent["class"]
            id, factory, lazy_init, abstract, parent, scope_ = self._get_basic_object_data(leaf, class_)

            c = self._create_object(id, factory, lazy_init, abstract, parent,
                           scope_, pos_constr, named_constr, props)

            self.objects.append(c)

        return parent

    def _get_pos_constr(self, object):
        """ Returns a list of all positional constructor arguments of an object.
        """
        if "constructor-args" in object and isinstance(object["constructor-args"], list):
            return [self._convert_prop_def(object, constr, object["object"]) for constr in object["constructor-args"]]
        return []

    def _get_named_constr(self, object):
        """ Returns a dictionary of all named constructor arguments of an object.
        """
        if "constructor-args" in object and isinstance(object["constructor-args"], dict):
            return dict([(name, self._convert_prop_def(object, constr, object["object"]))
                            for (name, constr) in object["constructor-args"].items()])
        return {}

    def _get_props(self, object):
        """ Returns a list of all properties defined by an object.
        """
        if "properties" in object:
            return [self._convert_prop_def(object, p, name) for (name, p) in object["properties"].items()]
        return []

    def _create_object(self, id, factory, lazy_init, abstract, parent,
                           scope, pos_constr, named_constr, props):
        """ A helper function which creates an object out of the supplied
        arguments.
        """

        c = ObjectDef(id=id, factory=factory, lazy_init=lazy_init,
            abstract=abstract, parent=parent)

        c.scope = scope
        c.pos_constr = pos_constr
        c.named_constr = named_constr
        c.props = props

        self.logger.debug("object: props = %s" % c.props)
        self.logger.debug("object: There are %s props" % len(c.props))

        return c

    def _get_basic_object_data(self, object, class_):
        """ A convenience method which creates basic object's data so that
        the code is not repeated.
        """

        if "scope" in object:
            scope_ = scope.convert(object["scope"])
        else:
            scope_ = scope.SINGLETON

        return(object["object"],  ReflectiveObjectFactory(class_),
            object.get("lazy-init", False), object.get("abstract", False),
               object.get("parent"), scope_)

    def _convert_object(self, object, prefix=""):
        "This function generates a object definition, then converts scope and property elements."
        if prefix != "":
            if "object" in object and object["object"] is not None:
                object["object"] = prefix + "." + object["object"]
            else:
                object["object"] = prefix + ".<anonymous>"

        id, factory, lazy_init, abstract, parent, scope_ = self._get_basic_object_data(object, object.get("class"))

        pos_constr = self._get_pos_constr(object)
        named_constr = self._get_named_constr(object)
        props = self._get_props(object)

        return self._create_object(id, factory, lazy_init, abstract, parent,
            scope_, pos_constr, named_constr, props)

    def _print_obj(self, obj, level=0):
        self.logger.debug("%sobject = %s" % ("\t"*level, obj["object"]))
        self.logger.debug("%sobject id = %s" % ("\t"*level, obj["object"]))
        self.logger.debug("%sclass = %s" % ("\t"*(level+1), obj["class"]))

        if "scope" in obj:
            self.logger.debug("%sscope = %s" % ("\t"*(level+1), obj["scope"]))
        else:
            self.logger.debug("%sscope = singleton (default)" % ("\t"*(level+1)))

        if "properties" in obj:
            self.logger.debug("%sproperties:" % ("\t"*(level+1)))
            for prop in obj["properties"].keys():
                if "object" in obj["properties"][prop]:
                    self.logger.debug("%s%s = ..." % ("\t"*(level+2), prop))
                    self._print_obj(obj["properties"][prop], level+3)
                else:
                    self.logger.debug("%s%s = %s" % ("\t"*(level+2), prop, obj["properties"][prop]))
        self.logger.debug("")

    def _convert_ref(self, ref_node, name):
        self.logger.debug("ref: Parsing %s, %s" % (ref_node, name))
        if "object" in ref_node:
            return ReferenceDef(name, ref_node["object"])
        else:
            return ReferenceDef(name, ref_node)

    def _convert_value(self, value, id, name):
        results = []

        if isinstance(value, dict):
            if "tuple" in value:
                self.logger.debug("value: Converting tuple")
                return self._convert_tuple(value["tuple"], id, name)
            elif "list" in value:
                self.logger.debug("value: Converting list")
                return self._convert_list(value["list"], id, name)
            elif "dict" in value:
                self.logger.debug("value: Converting dict")
                return self._convert_dict(value["dict"], id, name)
            elif "set" in value:
                self.logger.debug("value: Converting set")
                return self._convert_set(value["set"], id, name)
            elif "frozenset" in value:
                self.logger.debug("value: Converting frozenset")
                return self._convert_frozen_set(value["frozenset"], id, name)
        else:
            self.logger.debug("value: Plain ole value = %s" % value)
            return value

        return results

    def _convert_dict(self, dict_node, id, name):
        d = {}
        for (k, v) in dict_node.items():
            if isinstance(v, dict):
                self.logger.debug("dict: You have a special type stored at %s" % k)
                if "ref" in v:
                    self.logger.debug("dict/ref: k,v = %s,%s" % (k, v))
                    d[k] = self._convert_ref(v["ref"], "%s.dict['%s']" % (name, k))
                    self.logger.debug("dict: Stored %s => %s" % (k, d[k]))
                elif "tuple" in v:
                    self.logger.debug("dict: Converting a tuple...")
                    d[k] = self._convert_tuple(v["tuple"], id, "%s.dict['%s']" % (name, k))
                else:
                    self.logger.debug("dict: Don't know how to handle type %s" % v)
            else:
                self.logger.debug("dict: %s is NOT a dict, so going to convert as a value." % v)
                d[k] = self._convert_value(v, id, "%s.dict['%s']" % (name, k))
        return DictDef(name, d)

    def _convert_props(self, props_node, name):
        dict = {}
        for prop in props_node.prop:
            dict[prop.key] = str(prop)
        return DictDef(name, dict)

    def _convert_list(self, list_node, id, name):
        list = []
        for item in list_node:
            self.logger.debug("list: Adding %s to list..." % item)
            if isinstance(item, dict):
                if "ref" in item:
                    list.append(self._convert_ref(item["ref"], "%s.list[%s]" % (name, len(list))))
                elif "object" in item:
                    list.append(self._convert_inner_object(item, id, "%s.list[%s]" % (name, len(list))))
                elif len(set(["dict", "tuple", "set", "frozenset", "list"]) & set(item)) > 0:
                    list.append(self._convert_value(item, id, "%s.list[%s]" % (name, len(list))))
                else:
                    self.logger.debug("list: Don't know how to handle %s" % item.keys())
            else:
                list.append(item)
        return ListDef(name, list)

    def _convert_tuple(self, tuple_node, id, name):
        list = []
        self.logger.debug("tuple: tuple_node = %s, id = %s, name = %s" % (tuple_node, id, name))
        for item in tuple_node:
            if isinstance(item, dict):
                if "ref" in item:
                    list.append(self._convert_ref(item["ref"], name + ".tuple"))
                elif "object" in item:
                    list.append(self._convert_inner_object(item, id, "%s.tuple[%s]" % (name, len(list))))
                elif len(set(["dict", "tuple", "set", "frozenset", "list"]) & set(item)) > 0:
                    list.append(self._convert_value(item, id, "%s.tuple[%s]" % (name, len(list))))
                else:
                    self.logger.debug("tuple: Don't know how to handle %s" % item)
            else:
                list.append(item)
        return TupleDef(name, tuple(list))

    def _convert_set(self, set_node, id, name):
        s = set()
        self.logger.debug("set: set_node = %s, id = %s, name = %s" % (set_node, id, name))
        for item in set_node:
            if isinstance(item, dict):
                if "ref" in item:
                    s.add(self._convert_ref(item["ref"], name + ".set"))
                elif "object" in item:
                    s.add(self._convert_inner_object(item, id, "%s.set[%s]" % (name, len(s))))
                elif len(set(["dict", "tuple", "set", "frozenset", "list"]) & set(item)) > 0:
                    s.add(self._convert_value(item, id, "%s.set[%s]" % (name, len(s))))
                else:
                    self.logger.debug("set: Don't know how to handle %s" % item)
            else:
                s.add(item)
        return SetDef(name, s)

    def _convert_frozen_set(self, frozen_set_node, id, name):
        item = self._convert_set(frozen_set_node, id, name)
        self.logger.debug("frozenset: Just got back converted set %s" % item)
        self.logger.debug("frozenset: value is %s, which will be turned into %s" % (item.value, frozenset(item.value)))
        return FrozenSetDef(name, frozenset(item.value))

    def _convert_inner_object(self, object_node, id, name):
        self.logger.debug("inner object: Converting %s" % object_node)
        inner_object_def = self._convert_object(object_node, prefix="%s.%s" % (id, name))
        self.objects.append(inner_object_def)
        return InnerObjectDef(name, inner_object_def)

    def _convert_prop_def(self, comp, p, name):
        "This function translates object properties into useful collections of information for the container."
        self.logger.debug("prop_def: Trying to read property %s -> %s" % (name, p))
        if isinstance(p, dict):
            if "ref" in p:
                self.logger.debug("prop_def: >>>>>>>>>>>>Call _convert_ref(%s, %s)" % (p["ref"], name))
                return self._convert_ref(p["ref"], name)
            elif "tuple" in p:
                self.logger.debug("prop_def: Call _convert_tuple(%s,%s,%s)" % (p["tuple"], comp["object"], name))
                return self._convert_tuple(p["tuple"], comp["object"], name)
            elif "set" in p:
                self.logger.debug("prop_def: Call _convert_set(%s,%s,%s)" % (p["set"], comp["object"], name))
                return self._convert_set(p["set"], comp["object"], name)
            elif "frozenset" in p:
                self.logger.debug("prop_def: Call _convert_frozen_set(%s,%s,%s)" % (p["frozenset"], comp["object"], name))
                return self._convert_frozen_set(p["frozenset"], comp["object"], name)
            elif "object" in p:
                self.logger.debug("prop_def: Call _convert_inner_object(%s,%s,%s)" % (p, comp["object"], name))
                return self._convert_inner_object(p, comp["object"], name)
            else:
                #self.logger.debug("prop_def: Don't know how to handle %s" % p)
                return self._convert_dict(p, comp["object"], name)
        elif isinstance(p, list):
            return self._convert_list(p, comp["object"], name)
        elif isinstance(p, unicode):
            return ValueDef(name, unicode(p))
        else:
            return ValueDef(name, str(p))
        return None

        if hasattr(p, "ref"):
            return self._convert_ref(p.ref, name)
        elif hasattr(p, "value"):
            return ValueDef(name, str(p.value))
        elif hasattr(p, "dict"):
            return self._convert_dict(p.dict, comp.id, name)
        elif hasattr(p, "props"):
            return self._convert_props(p.props, name)
        elif hasattr(p, "list"):
            return self._convert_list(p.list, comp.id, name)
        elif hasattr(p, "tuple"):
            return self._convert_tuple(p.tuple, comp.id, name)
        elif hasattr(p, "set"):
            return self._convert_set(p.set, comp.id, name)
        elif hasattr(p, "frozenset"):
            self.logger.debug("Converting frozenset")
            return self._convert_frozen_set(p.frozenset, comp.id, name)
        elif hasattr(p, "object"):
            return self._convert_inner_object(p.object, comp.id, name)




"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
import logging
import threading
import Pyro.core, Pyro.naming

from socket import getaddrinfo, gethostbyname

pyro_threads = {}
serviceList = {}
logger = logging.getLogger("springpython.remoting.pyro.PyroDaemonHolder")

def resolve(host, port):
    canonhost = gethostbyname(host)
    canonport = getaddrinfo(host, port)[0][4][1]
    
    return canonhost, canonport

def register(pyro_obj, service_name, host, port):
    """
    Register the pyro object and its service name with the daemon.
    Also add the service to a dictionary of objects. This allows the
    PyroDaemonHolder to intelligently know when to start and stop the
    daemon thread.
    """
    logger.debug("Registering %s at %s:%s with the Pyro server" % (service_name, host, port))

    host, port = resolve(host, port)
    
    serviceList[(service_name, host, port)] = pyro_obj

    if (host, port) not in pyro_threads:
    
        logger.debug("Pyro thread needs to be started at %s:%s" % (host, port))
      
        pyro_threads[(host, port)] = _PyroThread(host, port)
        pyro_threads[(host, port)].start()
        
    pyro_threads[(host, port)].pyro_daemon.connect(pyro_obj, service_name)

def deregister(service_name, host, port):
    """
    Deregister the named service by removing it from the list of
    managed services and also disconnect from the daemon.
    """
    logger.debug("Deregistering %s at %s:%s with the Pyro server" % (service_name, host, port))

    host, port = resolve(host, port)
    
    pyro_threads[(host, port)].pyro_daemon.disconnect(serviceList[(service_name, host, port)])
    del(serviceList[(service_name, host, port)])

    def get_address((service_name, host, port)):
        return (host, port)

    if len([True for x in serviceList.keys() if get_address(x) == (host, port)]) == 0:
        logger.debug("Shutting down thread on %s:%s" % (host, port))
        shutdown(host, port)

def shutdown(daemon_host, daemon_port):
    """This provides a hook so an application can deliberately shutdown a
    daemon thread."""
    logger.debug("Shutting down pyro daemon at %s:%s" % (daemon_host, daemon_port))

    daemon_host, daemon_port = resolve(daemon_host, daemon_port)

    try:
        pyro_threads[(daemon_host, daemon_port)].shutdown()
        del(pyro_threads[(daemon_host, daemon_port)])
    except:
        logger.debug("Failed to shutdown %s:%s" % (daemon_host, daemon_port))

class _PyroThread(threading.Thread):
    """
    This is a thread that runs the Pyro daemon. It is instantiated automatically
    from within PyroServiceExporter.
    """
    
    def __init__(self, host, port):
        """
        When this class is created, it also created a Pyro core daemon to manage.
        """
        threading.Thread.__init__(self)
        self.host = host
        self.port = port
        self.logger = logging.getLogger("springpython.remoting.pyro.PyroDaemonHolder._PyroThread")

        self.pyro_daemon = Pyro.core.Daemon(host=host, port=port)
    
    def run(self):
        """
        When this thread starts up, it initializes the Pyro server and then puts the
        daemon into listen mode so it can process remote requests.
        """
        self._running = True
        self.logger.debug("Starting up Pyro server thread for %s:%s" % (self.host, self.port))
        Pyro.core.initServer()
        self.pyro_daemon.requestLoop(condition = lambda:self._running)

    def shutdown(self):
        """
        This is a hook in order to signal the thread that its time to shutdown
        the Pyro daemon.
        """
        self._running = False
        self.logger.debug("Signaling shutdown of Pyro server thread for %s:%s" % (self.host, self.port))





"""
   Copyright 2006-2008 SpringSource (http://springsource.com), All Rights Reserved

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.       
"""
from springpython.remoting.hessian.hessianlib import Hessian

class HessianProxyFactory(object):
    """
    This is wrapper around a Hessian client proxy. The idea is to inject this object with a 
    Hessian service_url, which in turn generates a Hessian client proxy. After that, any
    method calls or attribute accessses will be forwarded to the Hessian client proxy.
    """
    def __init__(self):
        self.__dict__["client_proxy"] = None
        
    def __setattr__(self, name, value):
        if name == "service_url":
            self.__dict__["service_url"] = value
        else:
            setattr(self.client_proxy, name, value)
    
    def __getattr__(self, name):
        if name == "service_url":
            return self.service_url
        elif name in ["post_process_before_initialization", "post_process_after_initialization"]:
            raise AttributeError, name
        else:
            if self.client_proxy is None:
                self.__dict__["client_proxy"] = Hessian(self.service_url)
            return getattr(self.client_proxy, name)

